import os
import copy
import pathlib
from . import attributes
from . import algorithms
from .util import hash
from .util import cache
from .util import commandline
from .util import materializer
from .util import encoder
from .util import parallel




def parseCameraPosition(text: str):
    parts = text.split (',')
    result = []
    if len (parts) in (6, 7):
        for p in parts:
            if re.match(r'^-?\d+(?:\.\d+)?$', p) is None:
                return None
            result.append (float (p))
        return result
    else:
        return None




def ParameterFilterCamera(param, value):
    result = parseCameraPosition (value)
    if result == None:
        return "invalid camera vector"
    return None




class _ToolRunner:
    """
    Class for building command line rendering tools.
    """
    def __init__ (self,
                    appName,
                    appGenerateScene,
                    appDescription,
                    appAnimators,
                    appCmdOptions,
                    appDefaultAttributes
        ):

        #: Name of the application.
        self.appName = appName

        #: Generator function for the scene.
        self.appGenerateScene = appGenerateScene

        #: Description of the application.
        self.appDescription = appDescription

        #: Animators supplied by application.
        self.appAnimators = appAnimators

        #: Additional command line options.
        self.appCmdOptions = appCmdOptions

        #: Default attributes from application
        self.appDefaultAttributes = appDefaultAttributes if appDefaultAttributes != None else attributes.Attributes.defaults ()

        # Create command line parser
        self.parser = commandline.Parser (self.appName, self.appDescription)

        # What to do?
        actions = {
            'duration':     {'description': 'Returns the animations duration in seconds.'},
            'generate':     {'description': 'Generate a script to be used by the materializer.'},
            'render':       {'description': 'Render an image.'},
            'generateseq':  {'description': 'Generate a sequence of scripts to be used together with materializer by running animators.'},
            'renderseq':    {'description': 'Render an image sequence by running animators.'},
            'animation':    {'description': 'Create animation as video file.'},
            'parts':        {'description': 'Create a list of maufacturable parts and their count.'},
            'manufacture':  {'description': 'Create maufacturing data for all parts or a single part.'},
        }
        self.argAction = commandline.ArgumentChoose ("action", "Action to perform:", actions)

        # Output file prefix
        self.argOutput = commandline.Argument ("output", "Prefix of output file name.", False, True)

        # Create command line options: rendering options
        renderingGroup = "Rendering options"
        self.optTime = commandline.Parameter (
            "time",
            "t",
            "Time of animation begin or animation time for a single frame.",
            renderingGroup,
            "0.0",
            "t",
            commandline.FilterNumber
        )

        self.optTimeScale = commandline.Parameter (
            "timescale",
            "s",
            "Scale factor for animation time.",
            renderingGroup,
            1,
            "factor",
            commandline.FilterNumberGreterZero
        )

        self.optDuration = commandline.Parameter (
            "duration",
            "d",
            "Duration of sequence or animation.",
            renderingGroup,
            None,
            "d",
            commandline.FilterNumber
        )

        self.optWidth = commandline.Parameter (
            "width",
            "w",
            "Width of image in pixels.",
            renderingGroup,
            "1920",
            "w",
            commandline.FilterIntegerGreterZero
        )

        self.optHeight = commandline.Parameter (
            "height",
            "h",
            "Height of image in pixels.",
            renderingGroup,
            "1080",
            "h",
            commandline.FilterIntegerGreterZero
        )

        self.optFps = commandline.Parameter (
            "fps",
            "f",
            "Number of frames per second, used for sequences and animations.",
            renderingGroup,
            "60.0",
            "fps",
            commandline.FilterNumberGreterZero
        )

        self.optAntialias = commandline.Parameter (
            "antialias",
            "a",
            "Anti aliasing setting in range 1..12, higher values will look smoother.",
            renderingGroup,
            "1",
            "val",
            commandline.FilterIntegerGreterZero
        )

        self.optQuality = commandline.Parameter (
            "quality",
            "q",
            "Render quality setting in range 1..12, higher values will look better.",
            renderingGroup,
            "1",
            "val",
            commandline.FilterInteger
        )

        renderMinAngle, renderMinSize, renderFixedCount, renderMinSlices = self.appDefaultAttributes.rasterizingAttributes ()

        self.optFa = commandline.Parameter (
            "fa",
            None,
            "Maximum angle difference between segments when rasterizing shapes and solids in degrees.",
            renderingGroup,
            renderMinAngle,
            "fa",
            commandline.FilterNumberGreterZero
        )

        self.optFs = commandline.Parameter (
            "fs",
            None,
            "Maximum distance error of segment when rasterizing shapes and solids in units.",
            renderingGroup,
            renderMinSize,
            "fs",
            commandline.FilterNumberGreterZero
        )

        # collect manufacture and render formats
        maunfactureRenderFormats = {}
        for m in materializer.getMaterializers ():
            for e in m.materializeExtensions ():
                if not e in maunfactureRenderFormats:
                    maunfactureRenderFormats[e] = {}
        for m in materializer.getMaterializers ():
            for e in m.renderExtensions ():
                if not e in maunfactureRenderFormats:
                    maunfactureRenderFormats[e] = {}


        self.optManufactureRenderType = commandline.ParameterChoose (
            "fmt", 
            "M",
            "Format for manufacturing or rendering files.",
            renderingGroup,
            maunfactureRenderFormats,
            None,
            "format"
        )

        # collect encoding presets from encoder registry
        encodingPresets = {}
        for encoderInstance in encoder.getEncoders ():
            for preset in encoderInstance.presets ():
                encodingPresets[encoderInstance.name () + '.' + str (preset)] = {}

        defaultEncodingFormat = None
        if len (encoder.getEncoders ()) > 0:
            enc = encoder.getEncoders ()[0]
            defaultEncodingFormat = enc.name () + '.' + enc.defaultPreset ()

        self.optEncode = commandline.ParameterChoose (
            "encoding", 
            "e", 
            "Video encoding preset, affects quality and encoding time", 
            renderingGroup, 
            encodingPresets, 
            defaultEncodingFormat,
            "preset"
        )

        self.optCamera = commandline.Parameter (
            "camera",
            "c",
            "Override camera vector.",
            renderingGroup,
            None,
            "view",
            ParameterFilterCamera
        )

        projections = {
            'perspective':          {'description' : 'Use perspective projection.'},
            'orthogonal':           {'description' : 'Use orthogonal projection.'}
        }

        self.optProjection = commandline.ParameterChoose (
            "projection",
            "p",
            "Projection type used for rendering.",
            renderingGroup,
            projections,
            None,
            "p"
        )

        self.optPreview = commandline.Flag (
            "preview",
            "P",
            "Preview frame while rendering",
            renderingGroup
        )

        self.optPartFilter = commandline.Parameter (
            "parts",
            'S',
            "Filter string for parts to manufacture.",
            renderingGroup,
            "*",
            "filter"
        )

        # Create command line options: miscellanous options
        miscGroup = "Miscellaneous options"
        self.optHelp = commandline.Flag (
            "help",
            None,
            "Show help.",
            miscGroup,
            self._showHelp
        )

        self.optCache = commandline.Parameter (
            "cache",
            None,
            "Path of cache directory.",
            miscGroup,
            ".pcsg.cache",
            "cache"
        )

        self.optJobs = commandline.Parameter (
            "jobs",
            "j",
            "Number of jobs to execute in parallel.",
            miscGroup,
            1,
            "jobs",
            commandline.FilterIntegerGreterZero
        )

        self.optNice = commandline.Parameter (
            "nice",
            "n",
            "Nice level [0..19] for process, higher nice level has lower priority.",
            miscGroup,
            "15",
            "nice",
            commandline.FilterInteger
        )

        self.optZeroOutput = commandline.Flag (
            "zero-output",
            "z",
            "Disable output of status.",
            miscGroup
        )


        # List of materializers
        materializers = {}
        for m in materializer.getMaterializers ():
            materializers[m.name ()] = {}
        self.optMaterializer = commandline.ParameterChoose ("materializer", "m", "Select materializer", miscGroup, materializers, "openscad", "m")

        # last progress display string
        self.lastProgress = ""


    def _registerToolOptions (self):
        # arguments
        self.parser.register (self.argAction)
        self.parser.register (self.argOutput)

        # rendering group
        self.parser.register (self.optTime)
        self.parser.register (self.optTimeScale)
        self.parser.register (self.optDuration)
        self.parser.register (self.optWidth)
        self.parser.register (self.optHeight)
        self.parser.register (self.optFps)
        self.parser.register (self.optQuality)
        self.parser.register (self.optAntialias)
        self.parser.register (self.optEncode)
        self.parser.register (self.optCamera)
        self.parser.register (self.optProjection)
        self.parser.register (self.optPreview)
        self.parser.register (self.optFa)
        self.parser.register (self.optFs)
        self.parser.register (self.optManufactureRenderType)
        self.parser.register (self.optPartFilter)

        # misc group
        self.parser.register (self.optHelp)
        self.parser.register (self.optMaterializer)
        self.parser.register (self.optCache)
        self.parser.register (self.optJobs)
        self.parser.register (self.optNice)
        self.parser.register (self.optZeroOutput)


    def _showHelp (self, *skip):
        """
        Show help an exit.
        """
        self.parser.showHelp ()
        os._exit (0)


    def _getDuration(self):
        """
        Get duration of animation.
        """
        maxDuration = 0
        if self.appAnimators != None:
            for animation in self.appAnimators:
                d = animation.getAnimationTime ()
                if d > maxDuration:
                    maxDuration = d
        return maxDuration


    def _doDuration(self):
        """
        Print duration of animation.
        """
        if self.optZeroOutput.value != True:
            print (str (self._getDuration ()))


    def _createJob (self, baseAttributes, materializer, time):
        """
        Creates a basic job
        """
        job = {}
        job['baseAttributes'] = baseAttributes  # basic attributes
        job['csg'] = None                       # storage for csg tree
        job['attributes'] = None                # resolved attributes
        job['csgHash'] = None                   # hash code of csg tree including attributes and calculated animations
        job['time'] = time                      # scene time
        job['animations'] = None                # precalculated animation positions 
        job['materializer'] = materializer      # materializer instance
        job['description'] = None               # job description for batch processing
        return job


    def _calculateAnimations (self, attributes, time):
        """
        Computes the animator positions for a render job.
        """
        positions = {}
        if self.appAnimators != None:
            for animation in self.appAnimators:
                positions["animations." + animation.name] = animation.getPosition (time)
        return positions


    def _createJobGenerate (self, baseAttributes, materializer, time, scriptFile = None):
        """
        Creates a generate job
        """
        job = self._createJob (baseAttributes, materializer, time)
        job['scriptCacheFile'] = None               # absolute path of generated cache file
        job['scriptFile'] = scriptFile              # target file name without extension, when output file should be copied
        job['description'] = "Generating " + materializer.name () + " script"
        job['runner'] = self._runJobGenerate        # runner function for job
        return job


    def _createJobRender (self, baseAttributes, materializer, time, extension, imageFile = None):
        """
        Creates a render job
        """
        job = self._createJob (baseAttributes, materializer, time)
        job['imageCacheFile'] = None                # absolute path of generated cache file
        job['imageFile'] = imageFile                # target file name without extension, when output file should be copied
        job['extension'] = extension                # extension for rendered file
        job['description'] = "Rendering " + extension + " image"
        job['runner'] = self._runJobRender          # runner function for job
        return job


    def _createJobEncode (self, baseAttributes, materializer, time, renderJobs, encoderInstance, encoderPreset, fps, animationFile = None):
        """
        Creates an encoding job
        """
        job = self._createJob (baseAttributes, materializer, time)
        job['renderJobs'] = renderJobs              # list of frame render jobs
        job['encoderInstance'] = encoderInstance    # encoder to use for video rendering
        job['encoderPreset'] = encoderPreset        # preset to be used for encoding
        job['fps'] = fps                            # target frames per second
        job['animationFile'] = animationFile        # target file name without extension, when output file should be copied
        job['description'] = "Encoding video"       # job description
        job['runner'] = self._runJobEncode          # runner function for job
        return job


    def _createJobManufacture (self, baseAttributes, materializer, part, partName, extension):
        """
        Creates a manufacturing job
        """
        job = self._createJob (baseAttributes, materializer, 0)
        job['part'] = part                          # part to manufacture
        job['partName'] = partName                  # file name of manufactured part
        job['description'] = "Manufacturing part"   # job description
        job['runner'] = self._runJobManufacture     # runner function for job
        job['extension'] = extension                # file extension for manufactured parts
        return job


    def _runJob (self, job):
        """
        Runnable for abstract job
        """
        # get attributes
        attrs = job['baseAttributes']

        # calculate animations
        if job['animations'] == None:
            job['animations'] = self._calculateAnimations (attrs, job['time'])

        # merge animators to attributes
        attrs = attrs.override (job['animations'])

        # merge camera settings
        cameraAttributes = attrs.cameraAttributes ()
        cameraView = cameraAttributes['view']
        cameraProjection = cameraAttributes['projection']

        # override camera attributes of scene from animations
        if 'animations.camera' in job['animations']:
            cameraView = job['animations']['animations.camera']

        # override camera settings from commandline
        attrs = attrs.override ({'camera.view': cameraView, 'camera.projection': cameraProjection})
        if self.optCamera.value != None:
            cameraView = parseCameraPosition (self.optCamera.value)
        if self.optProjection.value != None:
            cameraProjection = self.optProjection.value

        # set resolved attributes
        job['attributes'] = attrs

        # generate csg tree?
        if job['csg'] == None:
            job['csg'] = self.appGenerateScene (attrs)
            job['csgHash'] = hash (job['csg'])

        # okay, we are done.
        return True


    def _runJobGenerate (self, job):
        """
        Executes a generate job
        """
        # parent processing.
        if self._runJob (job) == False:
            return False

        # generate script and copy output file if requested
        job['scriptCacheFile'] = job['materializer'].script (job['csg'], job['attributes'], job['scriptFile'])
        return True


    def _runJobRender (self, job):
        """
        Executes a generate job
        """
        # parent processing.
        if self._runJob (job) == False:
            return False

        # generate script and copy output file if requested
        job['imageCacheFile'] = job['materializer'].render (job['csg'], job['attributes'], job['imageFile'], job['extension'])
        return True


    def _runJobEncode (self, job):
        """
        Executes an encode job
        """
        # parent processing.
        if self._runJob (job) == False:
            return False

        # collect frames and call encoder
        frames = []
        for renderJob in job['renderJobs']:
            frames.append (renderJob['imageCacheFile'])
        encoderInstance = job['encoderInstance']
        encoderPreset = job['encoderPreset']
        destPath = job['animationFile'] + '.' + encoderInstance.extension ()
        fps = job['fps']
        encoderInstance.encode (frames, fps, destPath, encoderPreset)
        return True


    def _runJobManufacture (self, job):
        """
        Runs a manufacturing job.
        """
        # parent processing.
        if self._runJob (job) == False:
            return False

        # get part details
        part = job['part']
        partName = job['partName']
        mat = job['materializer']
        attr = job['baseAttributes']
        extension = job['extension']

        # run materializer
        mat.materialize (part, attr, partName, extension)
        return True


    def _mapParts (self, baseAttributes, materializer):
        # create job to get a csg tree instance
        job = self._createJob (baseAttributes, materializer, 0)
        self._runJob (job)

        # collect parts
        parts = algorithms.CollectParts.execute (job['csg'], baseAttributes)
        collected = {}
        for part in parts:
            name = part.partName
            if name in collected:
                # check if part inside is the same
                assert part.children[0] == collected[name]['item'], "Two parts with the same name must have equal children."
                collected[name]['count'] += 1
            else:
                # just insert
                collected[name] = {'count': 1, 'item': part.children[0]}
        return collected


    @staticmethod
    def _filterMatchesPart (partName, filter):
        # check if part name is matched by filter
        if filter in ('', '*'):
            return True
        if filter[-1] == '*':
            if partName.startswith (filter[0:len(filter) - 1]):
                return True
        else:
            if partName == filter:
                return True
        return False


    def _filterPartName (self, partName):
        # query part filter from command line
        filters = self.optPartFilter.value.split (',')
        if filters == None:
            return True
        if len (filters) == 0:
            return True
        
        # check filters to match
        for f in filters:
            if _ToolRunner._filterMatchesPart (partName, f):
                return True

        # filter not matched
        return False


    @staticmethod
    def _encodePartName (partName):
        assert isinstance (partName, str)
        result = ""
        for ch in partName:
            if (ord (ch) >= ord ('a')) and (ord (ch) <= ord ('z')):
                result += ch
            elif (ord (ch) >= ord ('A')) and (ord (ch) <= ord ('Z')):
                result += ch
            elif (ord (ch) >= ord ('0')) and (ord (ch) <= ord ('9')):
                result += ch
            elif ch in ("-", "_", "."):
                result += ch
            elif ch in (" ", "\t", "\r", "\n"):
                result += "-"
            else:
                result += "_"
        return result


    def run (self, cmdArguments = None):
        """
        Run the command line tool.
        """
        # parse command line options
        self._registerToolOptions ()
        self.parser.parse (cmdArguments)

        # set nice level
        nice = 15
        nice = int (self.optNice.value)
        if nice < 0:
            nice = 0
        if nice > 19:
            nice = 19
        os.nice (nice)

        # initialize cache
        cache.setup (self.optCache.value)

        # run duration action?
        if self.argAction.value == "duration":
            self._doDuration ()
            return

        # check if materializer is valid
        mat = materializer.getMaterializer (self.optMaterializer.value)
        if mat == None:
            self.parser.commandlineInvalid ('Invalid materializer')
        matErr = mat.assertUsable ()
        if matErr != None:
            self.parser.commandlineInvalid ('Can not use materializer ' + self.optMaterializer.value + ': ' + matErr)

        # check if encoder is valid
        if self.argAction.value == 'animation':
            encoderInstance, encoderPreset = encoder.parseSettings (self.optEncode.value)
            if encoderInstance == None:
                self.parser.commandlineInvalid ("Invalid encoder settings.")
            encErr = encoderInstance.assertUsable ()
            if encErr != None:
                self.parser.commandlineInvalid ('Can not use video encoding preset ' + self.optEncode.value + ': ' + encErr)

        # setup basic attributes
        ovs = {}
        # rendering attributes
        if self.optWidth.value != None:
            ovs['render.width'] = int (self.optWidth.value)
        if self.optHeight.value != None:
            ovs['render.height'] = int (self.optHeight.value)
        if self.optQuality.value != None:
            # TODO: assert range in parser
            ovs['render.quality'] = int (self.optQuality.value)
        if self.optAntialias.value != None:
            # TODO: assert range in parser
            ovs['render.antialias'] = int (self.optAntialias.value)
        # rasterizing attributes
        if self.optFa.value != None:
            ovs['rasterize.minAngle'] = float (self.optFa.value)
        if self.optFs.value != None:
            ovs['rasterize.minSize'] = float (self.optFs.value)
        # create basic attributes
        baseAttributes = self.appDefaultAttributes.override (ovs)


        # get start time for rendering
        startTime = 0
        if self.optTime.value != None:
            startTime = float (self.optTime.value)

        # calculate duration for animations and sequences
        if self.optDuration.value != None:
            duration = float (self.optDuration.value)
        else:
            duration = self._getDuration () - startTime
            if duration < 0:
                duration = 0
        if duration < 0:
            self.parser.commandlineInvalid ("Duration must be positive.")

        # create sequence?
        if self.argAction.value in ('generateseq', 'renderseq', 'animation'):
            # create sequence
            sequence = []
            fps = float (self.optFps.value)
            if self.optTimeScale.value != None:
                timeScale = float (self.optTimeScale.value)
            else:
                timeScale = 1
            endTime = startTime + duration
            timePerFrame = 1.0 / (fps * timeScale)
            currentTime = startTime
            sequenceId = 0
            while (currentTime < endTime):
                sequence.append ((currentTime, "_%06d" % (sequenceId,)))
                currentTime += timePerFrame
                sequenceId += 1

            # render single frame if sequence is empty.
            if len (sequence) == 0:
                sequence = [(startTime, '')]
        else:
            # only generate a single frame
            sequence = [(startTime, '')]

        # run parts action?
        if self.argAction.value == "parts":
            if self.optZeroOutput.value != True:
                partMap = self._mapParts (baseAttributes, mat)
                for part in partMap:
                    partName = _ToolRunner._encodePartName (part)
                    print (str (partMap[part]['count']) + " " + partName)
            return

        # Ensure output is set
        if self.argOutput.value == None:
            self.parser.commandlineInvalid ("missing argument output")

        # initialize empty job queue
        jobQueue = []

        # run generate action?
        if self.argAction.value in ('generate', 'generateseq'):
            jobs = []
            for item in sequence:
                scriptFileName = str (pathlib.Path (self.argOutput.value + item[1]).resolve ())
                jobs.append (self._createJobGenerate (baseAttributes, mat, item[0], scriptFileName))
            jobQueue.append (jobs)

        # run render action?
        elif self.argAction.value in ('render', 'renderseq'):
            # check selected render format:
            renderFormatExtension = mat.renderDefaultExtension ()
            if self.optManufactureRenderType.value != None:
                renderFormatExtension = self.optManufactureRenderType.value
            
            # check if extension is supported
            if not renderFormatExtension in mat.renderExtensions ():
                self.parser.commandlineInvalid ("Can not use output format ." + renderFormatExtension + " to render images with " + mat.name () + ".")

            jobs = []
            for item in sequence:
                renderImageName = str (pathlib.Path (self.argOutput.value + item[1]).resolve ())
                jobs.append (self._createJobRender (baseAttributes, mat, item[0], renderFormatExtension, renderImageName))
            jobQueue.append (jobs)

        # run animation action?
        elif self.argAction.value == "animation":
            # using fixed rendering format for animations
            renderFormatExtension = 'png'

            # append rendering job for each frame
            renderJobs = []
            for item in sequence:
                renderJobs.append (self._createJobRender (baseAttributes, mat, item[0], renderFormatExtension, None))
            jobQueue.append (renderJobs)

            # animation output path without file extension
            animationOutput = str (pathlib.Path (self.argOutput.value))

            # create encoding job
            encJob = self._createJobEncode (baseAttributes, mat, startTime, renderJobs, encoderInstance, encoderPreset, fps, animationOutput)
            jobQueue.append ([encJob])

        # run manufacture action?
        elif self.argAction.value == "manufacture":
            # map parts
            partMap = self._mapParts (baseAttributes, mat)

            # get materialize format extension
            matFormatExtension = mat.materializeDefaultExtension ()
            if self.optManufactureRenderType.value != None:
                matFormatExtension = self.optManufactureRenderType.value
            
            # check if extension is supported
            if not matFormatExtension in mat.materializeExtensions ():
                self.parser.commandlineInvalid ("Can not use output format ." + matFormatExtension + " to materialize parts with " + mat.name () + ".")

            # filter parts
            partsToProcess = []
            for part in partMap:
                partName = _ToolRunner._encodePartName (part)
                if self._filterPartName (partName):
                    partsToProcess.append ((partName, partMap[part]))

            # create manufacture jobs
            manufactureJobs = []
            for part in partsToProcess:
                partFileName = str (pathlib.Path (self.argOutput.value)) + '_' + part[0]
                manufactureJobs.append (self._createJobManufacture (baseAttributes, mat, part[1]['item'], partFileName, matFormatExtension))

            # queue manufacture jobs
            jobQueue.append(manufactureJobs)

        # failed: action not handled.
        else:
            assert False, "Internal error: action not handled"

        # multi process job queues
        threadCount = int (self.optJobs.value)
        if threadCount < 1:
            threadCount = 1
        if threadCount > 64:
            threadCount = 64
        for block in jobQueue:
            pRunner = parallel.ParallelRunner (block, self._progessCallback)
            pRunner.execute (threadCount)
            self._endProgress ()

        # continue processing
        return True


    def _isConsole (self):
        # TODO: get on runtime
        # deactivate with --debug parameter 
        return True


    def _endProgress (self):
        """
        End of progress display.
        """
        if self.lastProgress != None:
            if self.optZeroOutput.value != True:
                if self._isConsole ():
                    print ("")
        self.lastProgress = None


    def _progessCallback (self, description, elementIndex = None, elementCount = None):
        """
        Diplay current progresss.
        """
        text = description
        if elementCount != None:
            if elementIndex != None:
                text = description + " (" + str (elementIndex) + " / " + str (elementCount) + ")"
        else:
            if elementIndex != None:
                text = description + " (" + str (elementIndex) + ")"
        text = "  " + text

        # TODO: implement displaying on console
        if not self._isConsole ():
            if self.optZeroOutput.value != True:
                print (text)
        else:
            llLen = len (self.lastProgress) if self.lastProgress != None else 0
            tlen = len (text)
            if tlen < llLen:
                text = text + " " * (llLen - tlen)
            if self.optZeroOutput.value != True:
                print (text + "\r", end = "")
            self.lastProgress = text




class Tool:
    """
    Base class for implementing own command line tools.
    """
    def __init__ (self,
                    name: str = None,
                    description: str = None
        ):

        #: Name of tool
        self.name = name

        #: Description of tool.
        self.description = description

        #: Command line options of tool.
        self.options = []

        #: Animators of tool.
        self.animators = []


    def addAnimation (self, animation):
        """
        Adds an animation to this tool.
        """
        self.animators.append (animation)


    def addOption (self, option):
        """
        Add a command line option to the tool.
        """
        self.options.append (option)


    def attributes (self):
        """
        Returns the default attributes to be applied on processing.
        """
        return attributes.Attributes.defaults ()


    def scene (self, attributes):
        """
        Generate a scene (csg tree).
        """
        assert False, "To be implemented by child class"


    def run (self, commandLineArgs: list = None):
        """
        Runs the tool.
        """
        t = _ToolRunner (self.name, self.scene, self.description, self.animators, self.options, self.attributes ())
        return t.run (commandLineArgs)



from . import tree
from .attributes import Attributes




class Transform:
    """
    Informational base class of :ref:`Transform<Transform>` objects. This empty class is implemented for type matching purposes.
    """
    def __init__ (self):
        #: The number of dimensions is determined from the child instance.      
        self.dimensions = None




class Translate(tree.Node, Transform):
    """
    Translates a :ref:`Shape<Shape>` or :ref:`Solid<Solid>`.

    A translation can be constructed by a vector of the parameters x, y, and z.

    When translating :ref:`Shapes<Shape>` only the x and y parameters must be used, z must be 0 (default value).
    """
    def __init__ (self,
                    children: object = None,
                    vector: tuple = None,
                    x: float = None,
                    y: float = None,
                    z: float = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert children are valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 1, "Expected exactly one children object."
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child."
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child."
            children = (children,)
        dimensions = children[0].dimensions
        if not isinstance (children[0], tree.Empty):
            assert dimensions in (2, 3), "Expected a two or three dimensional child element."
    
        super ().__init__ (children = children, name = name, attributes = attributes)

        # parameter assignments
        if vector != None:
            assert isinstance (vector, (list, tuple)), "Expected vector to be a translation vector."
            assert len (vector) in (0, 1, 2, 3), "Expected vector to be a translation vector."
            assert x == None, "Can not use parameters x and vector together."
            assert y == None, "Can not use parameters y and vector together."
            assert z == None, "Can not use parameters z and vector together."
            x = vector[0] if len (vector) > 0 else 0
            y = vector[1] if len (vector) > 1 else 0
            z = vector[2] if len (vector) > 2 else 0
        if x == None:
            x = 0
        if y == None:
            y = 0
        if z == None:
            z = 0
        assert isinstance (x, (int, float)), "Expected parameter x to be a number"
        assert isinstance (y, (int, float)), "Expected parameter y to be a number"
        assert isinstance (z, (int, float)), "Expected parameter z to be a number"

        # TODO: reactivate this check, currently breaks the resorting in openscad exporter.
#        if dimensions == 2:
#            assert z == 0, "When translating two dimensional objects, parameter z must be 0."

        #: Translation in direction of the x-axis.
        self.x = x

        #: Translation in direction of the y-axis.
        self.y = y

        #: Translation in direction of the z-axis.
        self.z = z

        # The number of dimensions is determined from the child instance.      
        self.dimensions = dimensions

        #: Tuple containing a single :ref:`Shape<Shape>` or :ref:`Solid<Solid>`.
        self.children = self.children


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
            'x',
            'y',
            'z'
        )




class Scale(tree.Node, Transform):
    """
    Scales a :ref:`Shape<Shape>` or :ref:`Solid<Solid>`.

    A scale can be constructed by a vector of the parameters sx, sy, and sz.

    When scaling :ref:`Shapes<Shape>` only the sx and sy parameters must be used, sz must be 1 (default value).
    """
    def __init__ (self,
                    children: object = None,
                    vector: tuple = None,
                    sx: float = None,
                    sy: float = None,
                    sz: float = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert children are valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 1, "Expected exactly one children object."
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child."
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child."
            children = (children,)
        dimensions = children[0].dimensions
        if not isinstance (children[0], tree.Empty):
            assert dimensions in (2, 3), "Expected a two or three dimensional child element."
    
        super ().__init__ (children = children, name = name, attributes = attributes)

        # parameter assignments
        if vector != None:
            assert isinstance (vector, (list, tuple)), "Expected vector to be a scale vector."
            assert len (vector) in (0, 1, 2, 3), "Expected vector to be a scale vector."
            assert sx == None, "Can not use parameters sx and vector together."
            assert sy == None, "Can not use parameters sy and vector together."
            assert sz == None, "Can not use parameters sz and vector together."
            sx = vector[0] if len (vector) > 0 else 1
            sy = vector[1] if len (vector) > 1 else 1
            sz = vector[2] if len (vector) > 2 else 1
        if sx == None:
            sx = 1
        if sy == None:
            sy = 1
        if sz == None:
            sz = 1
        assert isinstance (sx, (int, float)), "Expected parameter sx to be a number"
        assert isinstance (sy, (int, float)), "Expected parameter sy to be a number"
        assert isinstance (sz, (int, float)), "Expected parameter sz to be a number"

        if dimensions == 2:
            assert sz == 1, "When translating two dimensional objects, parameter sz must be 1."

        #: Scale in direction of the x-axis.
        self.sx = sx

        #: Scale in direction of the y-axis.
        self.sy = sy

        #: Scale in direction of the z-axis.
        self.sz = sz

        # The number of dimensions is determined from the child instance.      
        self.dimensions = dimensions

        #: Tuple containing a single :ref:`Shape<Shape>` or :ref:`Solid<Solid>`.
        self.children = self.children


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
            'sx',
            'sy',
            'sz'
        )




class Rotate(tree.Node, Transform):
    """
    Rotates a :ref:`Shape<Shape>` or :ref:`Solid<Solid>`.

    A rotation can be constructed by a vector of the parameters rx, ry, and rz.
    The rotation applies the 3 angles by rotating around the z axis, followed by a rotation around the y axis, finally a rotation around the x axis will be applied.

    When rotating :ref:`Shapes<Shape>` only the rz parameter is allowed to use, rx and ry must be 0 (default value).
    """
    def __init__ (self,
                    children: object = None,
                    vector: tuple = None,
                    rx: float = None,
                    ry: float = None,
                    rz: float = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):

        # assert children are valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 1, "Expected exactly one children object."
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child."
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child."
            children = (children,)
        dimensions = children[0].dimensions
        if not isinstance (children[0], tree.Empty):
            assert dimensions in (2, 3), "Expected a two or three dimensional child element."
    
        super ().__init__ (children = children, name = name, attributes = attributes)

        # parameter assignments
        if vector != None:
            assert isinstance (vector, (list, tuple)), "Expected vector to be a rotate vector."
            assert len (vector) in (0, 1, 2, 3), "Expected vector to be a rotate vector."
            assert rx == None, "Can not use parameters rx and vector together."
            assert ry == None, "Can not use parameters ry and vector together."
            assert rz == None, "Can not use parameters rz and vector together."
            rx = vector[0] if len (vector) > 0 else 0
            ry = vector[1] if len (vector) > 1 else 0
            rz = vector[2] if len (vector) > 2 else 0
        if rx == None:
            rx = 0
        if ry == None:
            ry = 0
        if rz == None:
            rz = 0
        assert isinstance (rx, (int, float)), "Expected parameter rx to be a number"
        assert isinstance (ry, (int, float)), "Expected parameter ry to be a number"
        assert isinstance (rz, (int, float)), "Expected parameter rz to be a number"

        if dimensions == 2:
            assert rx == 0, "When translating two dimensional objects, parameter rx must be 0."
            assert ry == 0, "When translating two dimensional objects, parameter ry must be 0."

        #: Rotation around the x-axis in degrees.
        self.rx = rx

        #: Rotation around the y-axis in degrees.
        self.ry = ry

        #: Rotation around the z-axis in degrees.
        self.rz = rz

        # The number of dimensions is determined from the child instance.      
        self.dimensions = dimensions

        #: Tuple containing a single :ref:`Shape<Shape>` or :ref:`Solid<Solid>`.
        self.children = self.children


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
            'rx',
            'ry',
            'rz'
        )



import os
import shutil
import pathlib
import atexit
import weakref
from . import random
from . import hash
from .. import immutable




class _Cache:
    """
    Cache class.
    """
    def __init__ (self, cachePrefix: str, tempPrefix: str, noPersistence: bool):
        # base path of cache directory
        if (cachePrefix == None) or (cachePrefix == ""):
            cachePrefix = ".pcsg.cache"

        #: random sequence, seeded with time stamp
        self.rand = random.RandomSequence ()

        #: cachePrefix supplied to constructor
        self.cachePrefix = cachePrefix

        #: absolute cache directory 
        self.cacheDir = pathlib.Path.absolute (pathlib.Path (cachePrefix + pathlib.os.sep + "cache"))
        if not self.cacheDir.exists ():
            try:
                self.cacheDir.mkdir (parents = True, exist_ok = True)
            except:
                pass

        #: tempPrefix supplied to constructor
        self.tempPrefix = tempPrefix

        #: location for temporary files
        if (tempPrefix == None) or (tempPrefix == ""):
            tempPrefix = cachePrefix + pathlib.os.sep + "temp" + pathlib.os.sep + self._createTempName ()

        #: flag to disable persistence
        self.noPersistence = noPersistence

        #: absolute temp directory, create if not existing
        self.tempDir = pathlib.Path.absolute (pathlib.Path (tempPrefix))
        tempCreated = False
        if not self.tempDir.exists ():
            tempCreated = True
            try:
                self.tempDir.mkdir (parents = True, exist_ok = True)
            except:
                tempCreated = False
        self._tempCreated = tempCreated

        # remove temp folder on application terminate
        atexit.register (self._onTerminate)

        #: weak cached items
        self._weakCache = weakref.WeakValueDictionary ()


    #: Alphabet to choose characters from when creating random temporary file names.
    _tempChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"


    def _createTempName (self, length = 12):
        """
        Generates a temporary name.
        """
        result = ""
        cid = 0
        while cid < length:
            rid = self.rand.intRange (len (_Cache._tempChars) - 1)
            rc = _Cache._tempChars[rid]
            result += rc
            cid += 1
        return result


    def _onTerminate (self):
        """
        Clean up temorary files on termination.
        """
        if self._tempCreated:
            try:
                shutil.rmtree (self.tempDir)
            except:
                pass


    def load (self, fingerprint, suffix):
        """
        Load instance from cache by fingerprint.
        """
        hashKey = '{:016x}'.format (hash ("path", fingerprint)) + "-" + str (suffix)
        cacheEntry = self._weakCache.get (hashKey)
        if cacheEntry != None:
            return cacheEntry
        if self.noPersistence:
            return None
        path = pathlib.Path (self.persistentPath (fingerprint, suffix))
        if path.exists:
            try:
                with path.open ("rb") as file:
                    obj = immutable.DataObject.restore (file, True)
                    self._weakCache[hashKey] = obj
                    return obj
            except:
                return None
        return None


    def store (self, instance, fingerprint, suffix):
        """
        Store instance by fingerprint to cache.
        """
        hashKey = '{:016x}'.format (hash ("path", fingerprint)) + "-" + str (suffix)
        self._weakCache[hashKey] = instance
        if self.noPersistence != True:
            pp = self.persistentPath (fingerprint, suffix, create = True)
            ppPre = pp + ".temp." + self._createTempName ()
            with open (ppPre, "wb") as file:
                instance.store (file)
            try:
                os.rename (ppPre, pp)
            except:
                pass
        return instance


    def persistentPath (self, fingerprint, suffix, create = False):
        """
        Create a persistent path for an item with a fingerprint
        """
        hashStr = '{:016x}'.format (hash ("path", fingerprint))
        hcPath = str (self.cacheDir) + pathlib.os.sep + hashStr[0:2] + pathlib.os.sep + hashStr[2:4]
        hcFile = hcPath + pathlib.os.sep + hashStr[4:len (hashStr)] + "." + suffix
        if create:
            if not pathlib.Path (hcPath).exists ():
                try:
                    pathlib.Path (hcPath).mkdir (parents = True, exist_ok = True)
                except:
                    pass
        return hcFile


    def temppath (self, suffix: str, create: bool = False):
        """
        Create temporary path name.
        """
        pathName = self.tempDir
        fileName = str (pathName) + pathlib.os.sep + self._createTempName ()
        if suffix != None:
            if suffix != "":
                fileName += "." + suffix
        if create:
            if not pathlib.Path (pathName).exists ():
                pathlib.Path (pathName).mkdir (parents = True, exist_ok = True)
        return fileName

"""
Single instance of cache.
"""
_CacheInstance = None




def setup (cachePrefix: str = None, tempPrefix: str = None, noPersistence: bool = False, ignoreIfAlreadySetup: bool = False):
    """
    Setup the cache module. The cache will operate on the directories specified by *cachePrefix* and *tempPrefix*.
    Setup must only be called once, when calling a second time with other parameters, an exception will be thrown.
    """
    global _CacheInstance
    if _CacheInstance != None:
        if not ignoreIfAlreadySetup:
            assert _CacheInstance.cachePrefix == cachePrefix, "Cache was already initialized with different arguments."
            assert _CacheInstance.tempPrefix == tempPrefix, "Cache was already initialized with different arguments."
            assert _CacheInstance.noPersistence == noPersistence, "Cache was already initialized with different arguments."
    else:
        _CacheInstance = _Cache (cachePrefix = cachePrefix, tempPrefix = tempPrefix, noPersistence = noPersistence)
    return _CacheInstance




def instance ():
    """
    Get single instance of cache. If not setup already, the cache will be set up by default parameters.
    """
    global _CacheInstance
    if _CacheInstance != None:
        return _CacheInstance
    else:
        return setup ()




def temporary (suffix: str = None):
    """
    Creates a temporary file name and returns it's absolute path.
    """
    cache = instance ()
    return cache.temppath (suffix)




def load (fingerprint, suffix: str = ""):
    """
    Load data object by fingerprint.
    """
    cache = instance ()
    return cache.load (fingerprint, suffix)




def store (item, fingerprint, suffix: str = ""):
    """
    Store data object by fingerprint.
    """
    cache = instance ()
    return cache.store (item, fingerprint, suffix)




def persistentPath (fingerprint, suffix: str = "", create: bool = False):
    """
    Get persistent absolute path of item by fingerprint.
    """
    cache = instance ()
    return cache.persistentPath (fingerprint, suffix, create)



from enum import Enum
from .util import vmath




class PathType(Enum):
    """
    Interpolation type of Path.
    """
    #: Stop animation. Length of the path must be zero.
    STOP = 0

    #: Stop the animation if path length is zero. Curve interpolation else.
    MAY_STOP = 1

    #: Create smooth curve as path.
    CURVE = 2

    #: Create straight line as path.
    LINEAR = 3

    #: Create step response.
    STEP = 4




class Path:
    """
    Path element of an Animation.
    """
    def __init__ (self, relativeTime = None, absolutePosition = None, pathType = PathType.CURVE, absoluteTime = None, relativePosition = None, rounding = None, damping = None):
        assert isinstance (pathType, PathType), "pathType is invalid"
        
        #: Interpolation type of Path.
        self.pathType = pathType

        # Begin position, assigned by Animation from predecessor. When None (default), the value will be taken from the parent Animation.
        self.beginPosition = None

        #: Rounding factor (range 0 to 1). Higher rounding factors look smoother.
        self.rounding = rounding
        if rounding != None:
            assert isinstance (rounding, (int, float)), "rounding must be a number"

        #: Acceleration damping factor (range 0 to 1). Higher damping factors look smoother. When None (default), the value will be taken from the parent Animation.
        self.damping = damping
        if damping != None:
            assert isinstance (damping, (int, float)), "damping must be a number"

        if absoluteTime != None:
            assert relativeTime == None, "not allowed to use absoluteTime together with relativeTime"
            assert isinstance (absoluteTime, (int, float)), "absoluteTime must be a number"
            _absTime = absoluteTime
            _relTime = None
        else:
            assert relativeTime != None, "absoluteTime or relativeTime must be provided"
            assert isinstance (relativeTime, (int, float)), "relativeTime must be a number"
            _absTime = None
            _relTime = relativeTime

        if absolutePosition != None:
            assert relativePosition == None, "not allowed to use absolutePosition together with relativePosition"
            _absPos = absolutePosition
            _relPos = None
        else:
            assert relativePosition != None, "absolutePosition or relativePosition must be provided"
            _absPos = None
            _relPos = relativePosition

        #: Absolute end time of this Path.
        self.absoluteTime = _absTime

        #: Duration of this Path.
        self.relativeTime = _relTime

        #: Absolute end position of this Path.
        self.absolutePosition = _absPos

        #: Relative end position of this Path.
        self.relativePosition = _relPos




class Animation:
    """
    An animation interpolates a control parameter between key frames. The interpolated value can be a vector or a float number.
    """
    def __init__(self, name: str, initialPosition: list, pathes: list = None, rounding: float = 1.0, damping: float = 1.0):
        assert initialPosition != None, "initialPosition must be provided"
        if vmath.isVector (initialPosition) == False:
            assert isinstance (initialPosition, (int, float)), "initialPosition must be a number or vector"

        #: Name of the animation. The name is used to create a key in the :ref:`Attributes<Attributes>` object.
        self.name = name

        #: The initial position of the animation as number or vector.
        self.initialPosition = initialPosition

        # List of pathes
        self.pathes = []

        #: Rounding factor (range 0 to 1). Higher rounding factors look smoother. Default is 1.
        self.rounding = rounding

        #: Acceleration damping factor (range 0 to 1). Higher damping factors look smoother. Default is 1.
        self.damping = damping

        # precomputed values
        self.curveCache = None

        # add pathes
        if pathes != None:
            for path in pathes:
                self.addPath (path)


    def getPosition(self, time):
        """
        Get animation position for a time stamp.
        """
        if time <= 0:
            return self.initialPosition
        if len (self.pathes) == 0:
            return self.initialPosition
        timeLeft = time
        for i in range (0, len (self.pathes)):
            if timeLeft <= self.pathes[i].relativeTime:
                return self._getPathPosition (i, timeLeft)
            timeLeft = timeLeft - self.pathes[i].relativeTime
        return self.getEndPosition ()


    def getAnimationTime(self):
        """
        Get time of animation.
        """
        if len (self.pathes) > 0:
            return self.pathes[len (self.pathes) - 1].absoluteTime
        return 0

    
    def getInitialPosition(self):
        """
        Get initial position of animation.
        """
        return self.initialPosition

    
    def getEndPosition(self):
        """
        Get end position of animation.
        """
        if len (self.pathes) > 0:
            return self.pathes[len (self.pathes) - 1].absolutePosition
        return self.initialPosition


    def path(self, relativeTime = None, absolutePosition = None, pathType = PathType.CURVE, absoluteTime = None, relativePosition = None, rounding = None, damping = None):
        """
        Convenience method to add pathes.
        """
        path = Path (relativeTime, absolutePosition, pathType, absoluteTime, relativePosition, rounding, damping)
        self.addPath (path)
        return self


    def addPath(self, path):
        """
        Adds a path to this animation.
        """
        assert isinstance (path, Path)
        if isinstance (self.initialPosition, (int, float)):
            if path.absolutePosition != None:
                assert isinstance (path.absolutePosition, (int, float)), "absolutePosition of path must match type of initialPosition"
            if path.relativePosition != None:
                assert isinstance (path.relativePosition, (int, float)), "relativePosition of path must match type of initialPosition"
        else:
            if path.absolutePosition != None:
                assert len (path.absolutePosition) == len (self.initialPosition), "absolutePosition of path must match type of initialPosition"
            if path.relativePosition != None:
                assert len (path.relativePosition) == len (self.initialPosition), "relativePosition of path must match type of initialPosition"
        if path.relativeTime == None:
            path.relativeTime = path.absoluteTime - self.getAnimationTime ()
        elif path.absoluteTime == None:
            path.absoluteTime = path.relativeTime + self.getAnimationTime ()
        if path.relativePosition == None:
            path.relativePosition = vmath.vsSub (path.absolutePosition, self.getEndPosition ())
        elif path.absolutePosition == None:
            path.absolutePosition = vmath.vsAdd (path.relativePosition, self.getEndPosition ())
        if path.pathType == PathType.MAY_STOP:
            if path.relativeTime <= 0:
                return
        assert path.relativeTime >= 0, "relativeTime of path must be >= 0"
        path.beginPosition = self.getEndPosition ()
        self.pathes.append (path)
        self._invalidateCaches ()


    def _invalidateCaches(self):
        """
        Animation changed, release cached elements.
        """
        self.curveCache = None


    def _getPathPredecessor(self, index, roundRunning = True):
        """
        Returns the index of a path's neigbour without stop pathes.
        """
        beginIndex = index
        candidate = index - 1
        if candidate < 0:
            candidate = len (self.pathes) - 1
        while candidate != beginIndex:
            path = self.pathes[candidate]
            if not path.pathType in (PathType.STOP, PathType.MAY_STOP):
                return candidate
            candidate = candidate - 1
            if candidate < 0:
                candidate = len (self.pathes) - 1
        return None


    def _getPathSuccessor(self, index, roundRunning = True):
        """
        Returns the index of a path's neigbour without stop pathes.
        """
        beginIndex = index
        candidate = index + 1
        if candidate >= len (self.pathes):
            candidate = 0
        while candidate != beginIndex:
            path = self.pathes[candidate]
            if not path.pathType in (PathType.STOP, PathType.MAY_STOP):
                return candidate
            candidate = candidate + 1
            if candidate >= len (self.pathes):
                candidate = 0
        return None


    def _calculateControlPoint(self, start, end, anchestorStart, anchestorEnd, rounding):
        """
        Compute a control point for a curve
        """
        ownLength = vmath.vsLength (vmath.vsSub (end, start))
        center = vmath.vsMulScalar (vmath.vsAdd (start, end), 0.5)
        direction = vmath.vsNormalize (vmath.vsSub (start, anchestorStart))
        distance = ownLength * rounding * 0.5
        return vmath.vsMulScalar (vmath.vsAdd (vmath.vsAdd (start, vmath.vsMulScalar (direction, distance)), center), 0.5)


    def _createCurve(self, index, predecessor, successor):
        """
        Create curve for a path.
        """
        # TODO: handle special cases like STOP, LINEAR and STEP
        # get local points
        path = self.pathes[index]
        begin = path.beginPosition
        end = path.absolutePosition
        # TODO: need two center points to emulate a line
        center = vmath.vsMulScalar (vmath.vsAdd (begin, end), 0.5)

        # get rounding value
        rounding = self.rounding
        if path.rounding != None:
            rounding = path.rounding

        # compute control points
        p2 = center
        p3 = center
        curved = path.pathType == PathType.CURVE
        if curved:
            # get points of adjacent pathes
            pre = self.pathes[predecessor] if predecessor != None else None
            succ = self.pathes[successor] if successor != None else None
            preBegin = pre.beginPosition if pre != None else None
            preEnd = pre.absolutePosition if pre != None else None
            succBegin = succ.beginPosition if succ != None else None
            succEnd = succ.absolutePosition if succ != None else None
            if (succBegin != None) and (succEnd != None):
                p3 = self._calculateControlPoint(end, begin, succEnd, succBegin, rounding)
            if (preBegin != None) and (preEnd != None):
                p2 = self._calculateControlPoint(begin, end, preBegin, preEnd, rounding)
        return [begin, p2, p3, end]


    def _getCurves(self):
        """
        Get or compute curves for animator.
        """
        if self.curveCache != None:
            return self.curveCache
        curves = []
        roundRunning = True
        for index in range (0, len (self.pathes)):
            predecessor = self._getPathPredecessor (index, roundRunning)
            successor = self._getPathSuccessor (index, roundRunning)
            curves.append (self._createCurve (index, predecessor, successor))
        self.curveCache = curves
        return curves


    def _getPathPosition(self, pathId, timeInPath):
        """
        Get animation position inside path.
        """
        path = self.pathes[pathId]
        relTime = 0
        if path.relativeTime > 0.0001:
            relTime = timeInPath / path.relativeTime
        if path.pathType == PathType.CURVE:
            # bezier interpolation
            curves = self._getCurves ()
            return vmath.bezier (curves[pathId], relTime)
        elif path.pathType == PathType.STOP:
            return path.absolutePosition
        elif path.pathType == PathType.MAY_STOP:
            return path.absolutePosition
        elif path.pathType == PathType.STEP:
            return path.absolutePosition
        else:
            # linear interpolation
            return vmath.vsAdd (vmath.vsMulScalar (path.absolutePosition, relTime), vmath.vsMulScalar (path.beginPosition, 1.0 - relTime))



import inspect
import pickle
from . import util




class Object:
    """
    Base class for immutable objects: only allows attributes to be assigned from the constructor.
    Assigment of attributes outside the constructor will be prohibited.
    """
    def __init__ (self):
        pass


    def __setattr__ (self, name, value):
        """
        Prevent assigning attributes outside from own constructor as tree nodes are immutable.
        """
        stackframe = inspect.stack()[1]
        message = "Can not assign attribute: object is immutable"
        assert stackframe.function in ("__init__", "__setstate__"), message
        assert stackframe[0].f_locals["self"].__class__ == self.__class__, message
        super ().__setattr__ (name, value)


    def __delattr__ (self, name):
        """
        Prevent deleting attributes outside from own constructor as tree nodes are immutable.
        """
        stackframe = inspect.stack()[1]
        message = "Can not assign attribute: object is immutable"
        assert stackframe.function == "__init__", message
        assert stackframe[0].f_locals["self"].__class__ == self.__class__, message
        super ().__delattr__ (name)




class DataObject (Object):
    """
    Base class implementing copy construction, hash and comparison functions for immutable data structures.

    Each derived class must define a static method *_copyConstructorAttributes ()* returning a tuple containing base classes and field names.
    The returned tuple defines which attributes are supplied to the constructor when using the *copy ()* function.

    Each derived class could implement a static method *_compareAttributes ()* to select the fields used for hash and comparision functions.
    When this method is not implemented, *_copyConstructorAttributes ()* will be used instead.
    """
    def __init__ (self):
        super ().__init__ ()

        #: Class hash to check if a persisted objects class has been changed.
        self._restore_class_hash_ = None

        #: Cache dictionary for precomputed values.
        self._cache = {}


    def store (self, file):
        """
        Store *DataObject* to file, serializes all attributes defined by *_compareAttributes ()*.
        """
        pickle.dump (self, file)


    @staticmethod
    def restore (file, noException: bool = True):
        """
        Restore *DataObject* from file, calls the copy constructor after deserialization.
        When *noException* is True, None will be returned on failed restoration, othwerwise an exception will be thrown.
        Restoration will fail if the implementation has been changed.
        """
        if noException:
            try:
                return pickle.load (file)
            except:
                return None
        else:
            return pickle.load (file)


    def copy (self, **assignments):
        """
        Creates a copy of this object with altered parameters.
        All named parameters of the constructor can be applied to this method,
        when a parameter of the constructor is not supplied, the parameter will be
        inherited from the current instance.
        """
        params = {}
        attributes = DataObject._class_get_copy_fields (self.__class__)
        for attribute in attributes:
            params[attribute] = getattr (self, attribute, None)
        for key in assignments:
            params[key] = assignments[key]
        return self.__class__ (**params)


    def __hash__ (self):
        """
        Returns the hash code of this object.
        """
        if 'hash' in self._cache:
            return self._cache['hash']
        h = self._calculateHash ()
        self._cache['hash'] = h
        return h


    def __getstate__ (self):
        """
        Get attributes for persisting item.
        """
        params = {'_restore_class_hash_': util.hash (self.__class__)}
        attributes = DataObject._class_get_compare_fields (self.__class__)
        for attribute in attributes:
            params[attribute] = getattr (self, attribute, None)
        return params

    
    def __setstate__ (self, state):
        """
        Restore object state after reading persisted item.
        """
        self.__dict__ = state
        if self._restore_class_hash_ != util.hash (self.__class__):
            assert False, "Implement proper exception: persisted class doesn't match own implementation."
        params = {}
        attributes = DataObject._class_get_compare_fields (self.__class__)
        for attribute in attributes:
            params[attribute] = getattr (self, attribute, None)
        self.__init__ (**params)


    def _calculateHash (self):
        """
        Default implementation for computing the hash of a tree element. This function depends on the classes
        *_compareAttributes ()* or *_copyConstructorAttributes()* declarations.
        """
        clz = self.__class__
        fields = DataObject._class_get_compare_fields (clz)
        h = str (util.hash (clz)) + "@"
        for field in fields:
            h = h + "," + str (util.hash (getattr (self, field, None)))
        return util.hash (h)


    @staticmethod
    def _class_collect_compare_fields (clz, fields):
        # declares own _copyConstructorAttributes() ?
        attributes = None
        m = getattr (clz, '_compareAttributes', None)
        if m != None:
            mq = m.__qualname__
            if mq == (clz.__qualname__ + '._compareAttributes') and (clz.__module__ == m.__module__):
                attributes = m ()
        if attributes == None:
            m = getattr (clz, '_copyConstructorAttributes', None)
            if m != None:
                mq = m.__qualname__
                if mq == (clz.__qualname__ + '._copyConstructorAttributes') and (clz.__module__ == m.__module__):
                    attributes = m ()
        # process declared class attributes if supplied
        if attributes != None:
            for attribute in attributes:
                if inspect.isclass (attribute):
                    DataObject._class_collect_compare_fields (attribute, fields)
                elif isinstance (attribute, str):
                    if not attribute in fields:
                        fields.append (attribute)
                else:
                    assert False, "Unexpected type in _copyConstructorAttributes"
            return
        # fallback: collect from base classes
        for base in clz.__bases__:
            DataObject._class_collect_compare_fields (base, fields)


    @staticmethod
    def _class_collect_copy_fields (clz, fields):
        # declares own _copyConstructorAttributes() ?
        attributes = None
        className = clz.__qualname__
        m = getattr (clz, '_copyConstructorAttributes', None)
        if m != None:
            mq = m.__qualname__
            if mq == (clz.__qualname__ + '._copyConstructorAttributes') and (clz.__module__ == m.__module__):
                attributes = m ()
        # process declared class attributes if supplied
        if attributes != None:
            for attribute in attributes:
                if inspect.isclass (attribute):
                    DataObject._class_collect_copy_fields (attribute, fields)
                elif isinstance (attribute, str):
                    if not attribute in fields:
                        fields.append (attribute)
                else:
                    assert False, "Unexpected type in _copyConstructorAttributes"
            return
        # fallback: collect from base classes
        for base in clz.__bases__:
            DataObject._class_collect_copy_fields (base, fields)


    @staticmethod
    def _class_get_compare_fields (clz):
        """
        Returns a pre-computed list of field names to be used for hash and equality operations.
        """
#        cache = getattr (clz, '_cache_compare_fields_', None)
        # TODO: fix cache, this is not threadsafe!
#        if cache != None:
#            return cache
        fields = []
        DataObject._class_collect_compare_fields (clz, fields)
#        setattr (clz, '_cache_compare_fields_', fields)
        return fields


    @staticmethod
    def _class_get_copy_fields (clz):
        """
        Returns a pre-computed list of field names to be used to supply to the copy constructor.
        """
#        cache = getattr (clz, '_cache_copy_fields_', None)
        # TODO: fix cache, this is not threadsafe!
#        if cache != None:
#            return cache
        fields = []
        DataObject._class_collect_copy_fields (clz, fields)
#        setattr (clz, '_cache_copy_fields_', fields)
        return fields



import hashlib
import copy
import datetime




class RandomSequence:
    """
    Pseudo random generator.
    """
    def __init__(self, seed = None):
        self.state = hashlib.md5 ()
        if seed != None:
            self.state.update (str (seed).encode ('utf-8'))
        else:
            self.state.update (str (datetime.datetime.now ().utctimetuple ()).encode ('utf-8'))
        self.counter = 0


    def _next(self):
        """
        Get next pseudo random entry MD5.
        """
        self.state.update (str (self.counter).encode ('utf-8'))
        self.counter = self.counter + 17
        return str (self.state.hexdigest ())


    def _nextInteger(self, signed = False):
        """
        Get next 64 bit signed or unsigned integer of random sequence.
        """
        nextVal = self._next ()
        intNum = int ("0x" + (nextVal [0 : 16]), base = 16)
        if signed == False:
            return intNum
        if (intNum & 0x80000000) != 0:
            intNum = 0xffffffff - (intNum + 1)
        return intNum


    def intRange (self, max, min = 0):
        """
        Generate a pseudo random integer in range, max is excluded.
        """
        assert max > min
        dist = max - min
        r = self._nextInteger (False)
        return min + (r % dist)


    def range (self, max = 1.0, min = 0.0):
        """
        Generate a pseudo random float in range [min..max], max is excluded.
        """
        fv = self._nextInteger () * 5.421010861e-20
        assert max > min
        dist = max - min
        return (fv * dist)  + min



import math




def isVector (any, expectedSize:int = None) -> bool:
    """
    Check if a data type is a valid vector containing only numbers. Optionally, the number of elements can be expected.
    """
    if isinstance (any, (list, tuple)):
        if expectedSize != None:
            if len (any) != expectedSize:
                return False
        for item in any:
            if not ((type (item) == int) or (type (item) == float)):
                return False
        return True
    return False




def assertVector (any, expectedSize: int = None, title: str = None, msg: str = None) -> bool:
    """
    Assert expression is a vector containing only numbers. Optionally, the number of elements can be expected.
    """
    if title == None:
        assert isVector (any), "expected a vector."
    else:
        if msg == None:
            assert isVector (any), title + " must be a vector."
        else:
            assert isVector (any), msg




def isNumber (any):
    """
    Check if data type is a number.
    """
    return (type (any) == int) or (type (any) == float)




def assertNumber (any, title: str = None, msg: str = None) -> None:
    """
    Asserts expression is a number.
    """
    if title == None:
        assert isNumber (any), "expected a number."
    else:
        if msg == None:
            assert isNumber (any), title + " must be a number."
        else:
            assert isNumber (any), msg



def vsAdd (vs1, vs2):
    """
    Add two scalars or vectors with same signature.
    """
    if isinstance (vs1, (int, float)):
        assert isinstance (vs2, (int, float)), "types must match"
        return vs1 + vs2
    assertVector (vs1)
    assertVector (vs2, len (vs1))
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] + vs2[i])
    return result




def vsAddScalar (vs1, vs2):
    """
    Add scalar or vector to scalar value.
    """
    assert isinstance (vs2, (int, float)), "second value must be a number"
    if isinstance (vs1, (int, float)):
        return vs1 + vs2
    assertVector (vs1)
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] + vs2)
    return result




def vsSub (vs1, vs2):
    """
    Difference of two scalars or vectors with same signature.
    """
    if isinstance (vs1, (int, float)):
        assert isinstance (vs2, (int, float)), "types must match"
        return vs1 - vs2
    assertVector (vs1)
    assertVector (vs2, len (vs1))
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] - vs2[i])
    return result




def vsSubScalar (vs1, vs2):
    """
    Difference of scalar or vector and scalar value.
    """
    assert isinstance (vs2, (int, float)), "second value must be a number"
    if isinstance (vs1, (int, float)):
        return vs1 - vs2
    assertVector (vs1)
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] - vs2)
    return result




def vsMul (vs1, vs2):
    """
    Product of two scalars or vectors with same signature.
    """
    if isinstance (vs1, (int, float)):
        assert isinstance (vs2, (int, float)), "types must match"
        return vs1 * vs2
    assertVector (vs1)
    assertVector (vs2, len (vs1))
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] * vs2[i])
    return result




def vsMulScalar (vs1, vs2):
    """
    Product of scalar or vector and scalar value.
    """
    assert isinstance (vs2, (int, float)), "second value must be a number"
    if isinstance (vs1, (int, float)):
        return vs1 * vs2
    assertVector (vs1)
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] * vs2)
    return result




def vsDiv (vs1, vs2):
    """
    Quotient of two scalars or vectors with same signature.
    """
    if isinstance (vs1, (int, float)):
        assert isinstance (vs2, (int, float)), "types must match"
        return vs1 / vs2
    assertVector (vs1)
    assertVector (vs2, len (vs1))
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] / vs2[i])
    return result




def vsDivScalar (vs1, vs2):
    """
    Quotient of scalar or vector and scalar value.
    """
    assert isinstance (vs2, (int, float)), "second value must be a number"
    if isinstance (vs1, (int, float)):
        return vs1 / vs2
    assertVector (vs1)
    result = []
    for i in range (0, len (vs1)):
        result.append (vs1[i] / vs2)
    return result




def vsLength (vs):
    """
    Get length of vector.
    """
    if isinstance (vs, (int, float)):
        return vs if vs >= 0 else -vs
    else:
        assertVector (vs)
        accu = 0
        for component in vs:
            accu = accu + component * component
        return math.sqrt (accu)




def vsNormalize (vs):
    """
    Normalize a vector.
    """
    length = vsLength (vs)
    if length == 0:
        return vs
    return vsDivScalar (vs, length)




def vsScalarProduct (vs1, vs2):
    """
    Scalar product of two vectors with same signature.
    """
    if isinstance (vs1, (int, float)):
        assert isinstance (vs2, (int, float)), "types must match"
        return vs1 * vs2
    assertVector (vs1)
    assertVector (vs2, len (vs1))
    accumulator = 0
    for component in range (0, len (vs1)):
        accumulator += vs1[component] * vs2[component]
    return accumulator




def calculateAngle (vs1, vs2):
    """
    Calculate the angle between two vectors.
    Returns the angle as radiant.
    """
    ma = vsScalarProduct (vs1, vs2) / (vsLength (vs1) * vsLength (vs2))
    if ma > 1:
        ma = 1
    return math.acos (ma)




def distanceLinePoint (lineP1, lineP2, point):
    """
    Calculates the distance between a point and a line defined by two points.
    """
    assertVector (lineP1)
    assertVector (lineP2, len (lineP1))
    assertVector (point, len (lineP1))
    nominator = 0
    denominator = 0
    for component in range (0, len (lineP1)):
        nominator -= lineP2[component] * point[component]
        nominator += lineP1[component] * lineP2[component]
        denominator -= lineP2[component] * lineP2[component]
    s = nominator / denominator
    distSquared = 0
    for component in range (0, len (lineP1)):
        distComponent = point[component] - lineP1[component] - s * lineP2[component]
        distSquared += distComponent * distComponent
    return math.sqrt (distSquared)




def bezier (points, t):
    """
    Calculate point of bezier curve.
    """
    t1 = (1 - t) * (1 - t) * (1 - t)
    t2 = 3 * t * (1 - t) * (1 - t)
    t3 = 3 * t * t * (1 - t)
    t4 = t * t * t
    return vsAdd (vsAdd(vsMulScalar (points[0], t1), vsMulScalar (points[1], t2)), vsAdd(vsMulScalar (points[2], t3), vsMulScalar (points[3], t4)))



import threading




class ParallelRunner:
    """
    Parallel executor for jobs
    """
    def __init__(self, jobs, progressCallback = None):
        self.lock = threading.Lock ()
        self.jobs = jobs
        self.progressCallback = progressCallback
        self.nextJob = 0
        self.executedJobs = 0

    
    def execute(self, threadCount):
        """
        Execute job list multi threaded.
        """
        threads = []
        # start worker threads
        for i in range (0, threadCount):
            t = threading.Thread (target = self._entry)
            threads.append (t)
            t.start ()

        # join worker threads
        for t in threads:
            t.join ()
            
        # end if sequence
        if self.progressCallback != None:
            lc = self.jobs[0]["description"] if len (self.jobs) > 0 else None
            if lc != None:
                self.progressCallback (lc, len (self.jobs), len (self.jobs))
        return True


    def _entry(self):
        """
        Thread entry function.
        """
        while self._doWork ():
            pass


    def _doWork(self) -> bool:
        """
        Process a single job. Returns true if there is more work.
        """
        # grab a job from queur
        self.lock.acquire ()
        jobId = self.nextJob
        self.nextJob = jobId + 1
        self.lock.release ()

        # abort if no jobs are left
        if jobId >= len (self.jobs):
            return False

        # execute job
        job = self.jobs[jobId]
        lc = job["description"]
        if self.progressCallback != None:
            self.lock.acquire ()
            self.progressCallback (lc, self.executedJobs, len (self.jobs))
            self.lock.release ()
        else:
            print (lc + " (" + str (jobId) + " / " + str (len (self.jobs)) + ")\n")
        if job["runner"] (job) == False:
            return False
        self.executedJobs = self.executedJobs + 1

        # continue on jobs
        return True




def execute (jobs: list, threadCount: int, progressCallback = None):
    """
    Execute workload in a thread pool.
    """
    # execute multi threaded
    if threadCount > 1:
        runner = ParallelRunner (jobs, progressCallback)
        return runner.execute (threadCount)

    # execute single threaded
    jid = 0
    jc = len (jobs)
    lc = None
    for job in jobs:
        lc = job["description"]
        if progressCallback != None:
            progressCallback (lc, jid, jc)
        else:
            print (lc + " (" + str (jid) + " / " + str (jc) + ")")
        if job["runner"] (job) == False:
            return False
        jid = jid + 1
    if jc > 0:
        if progressCallback != None:
            progressCallback (lc, jc, jc)
        else:
            print (lc + " (" + str (jc) + " / " + str (jc) + ")")
    return True



import subprocess
import pathlib
from shutil import copyfile
from .external import openscad




class Materializer:
    """
    Base class of materializer used to render an image from a csg tree or to materialize them into a 3D format.
    """

    def __init__ (self):
        pass


    def name (self):
        """
        Returns the name of the materializer.
        """
        assert False, "To be implemented by child"


    def assertUsable (self):
        """
        Returns None if the materializer is usable, an error string if not.
        """
        assert False, "To be implemented by child"


    def scriptExtension (self):
        """
        Returns the script path extension of the materializer.
        """
        assert False, "To be implemented by child"


    def script (self, csgTree, attributes = None, destinationPath = None):
        """
        Creates a script for materializing or rendering.
        """
        assert False, "To be implemented by child"


    def renderExtensions (self):
        """
        Returns a list of file extensions supported for rendering. When the materializer supports no rendering, the list will be empty.
        """
        assert False, "To be implemented by child"


    def renderDefaultExtension (self):
        """
        Returns the default file extension used for rendering.
        """
        assert False, "To be implemented by child"


    def render (self, csgTree, attributes = None, destinationPath = None, extension = None):
        """
        """
        assert False, "To be implemented by child"


    def materializeExtensions (self):
        """
        Returns a list of file extensions supported for rendering. When the materializer supports no rendering, the list will be empty.
        """
        assert False, "To be implemented by child"


    def materializeDefaultExtension (self):
        """
        Returns the default file extension used for materializing.
        """
        assert False, "To be implemented by child"


    def materialize (self, csgTree, attributes = None, destinationPath = None, extension = None):
        """
        """
        assert False, "To be implemented by child"




class _OpenScadMaterializer (Materializer):
    """
    Materializer using OpenScad.
    """
    def __init__ (self):
        super ().__init__ ()
        self.usableChecked = False
        self.usableError = None


    def name (self):
        return "openscad"


    def assertUsable (self):
        if self.usableChecked:
            return self.usableError
        try:
            result = subprocess.run ([openscad.getOpenScadRunnable (), '--version'], capture_output = True)
            if result.returncode != 0:
                self.usableError = "Can not execute openscad. Is inside PATH variable?"
            else:
                self.usableError = None
        except:
            self.usableError = "Can not execute openscad. Is inside PATH variable?"
        self.usableChecked = True
        return self.usableError
        

    def scriptExtension (self):
        return "scad"


    def script (self, csgTree, attributes = None, destinationPath = None):
        cacheFile = openscad.getScadScript (csgTree, attributes)
        if cacheFile != None:
            if destinationPath != None:
                dstFile = destinationPath + '.' + self.scriptExtension ()
                copyfile (cacheFile, dstFile)
                return dstFile
        return cacheFile


    def renderExtensions (self):
        return ('png', 'svg')


    def renderDefaultExtension (self):
        return 'png'
        

    def render (self, csgTree, attributes = None, destinationPath = None, extension = None):
        # TODO: check valid extension
        if extension == None:
            extension = self.renderExtensions () [0]
        cacheFile = openscad.getRendering (extension, csgTree, attributes)
        if cacheFile != None:
            if destinationPath != None:
                dstFile = destinationPath + '.' + extension
                copyfile (cacheFile, dstFile)
                return dstFile
        return cacheFile

    
    def materializeExtensions (self):
        return ('stl', 'off')


    def materializeDefaultExtension (self):
        return 'stl'


    def materialize (self, csgTree, attributes = None, destinationPath = None, extension = None):
        # TODO: check valid extension
        if extension == None:
            extension = self.materializeExtensions () [0]
        cacheFile = openscad.getMaterialized (extension, csgTree, attributes)
        if cacheFile != None:
            if destinationPath != None:
                dstFile = destinationPath + '.' + extension
                copyfile (cacheFile, dstFile)
                return dstFile
        return cacheFile




#: Static instance of materializer using OpenScad.
OpenScad = _OpenScadMaterializer ()




# list of registered materializers
_registeredMaterializers = [OpenScad]




def getMaterializers ():
    """
    Returns a list of all registered materializer instances.
    """
    return _registeredMaterializers




def getMaterializer (name):
    """
    Returns a materializer by name or None if not found.
    """
    for m in _registeredMaterializers:
        if m.name () == name:
            return m
    return None




def registerMaterializer (materializer):
    """
    Registers a materializer at the registry.
    """
    assert False, "TODO:"



import inspect
import enum




def _hashString (text):
    """
    Compute hash for a string, returns an unsigned 64 bit number.
    """
    value = 0x35af7b2c97a78b9e
    step = 0x072f2b592a4c57f9
    for ch in text:
        value = ((value * 104297) + (step * ord (ch))) & 0xFFFFFFFFFFFFFFFF
    return value




def _hashDict (element):
    """
    Compute hash for a dictionary, returns an unsigned 64 bit number.
    """
    value = "dict"
    for key in sorted(element):
        value = value + ":" + key + "=" + str (_hashArgument (element[key]))
    return _hashString (value)




def _hashList (element):
    """
    Compute hash for a list, returns an unsigned 64 bit number.
    """
    value = "list"
    for e in element:
        value = value + "," + str (_hashArgument (e))
    return _hashString (value)




def _hashObject (element):
    """
    Generate hash for an object, returns an unsigned 64 bit number.
    """
    m = getattr (element, "__hash__", None)
    if (m != None) and (m.__qualname__ != 'object.__hash__'):
        return element.__hash__ ()
    value = str (getattr (element, "__class__", "None")) + "@"
    for memberName in dir (element):
        member = getattr (element, memberName)
        if not memberName in ("__class__", "__dict__", "__doc__", "__module__", "__weakref__", "__slotnames__"):
            if not callable (member):
                value = "!" + memberName + "=" + str (_hashArgument (member))
    return _hashString (value)




_hashClassCache = {}
def _hashClass (element):
    """
    Generate hash for a class, returns an unsigned 64 bit number.
    Hashes of classes will track also code changes.
    """
    if element in _hashClassCache:
        return _hashClassCache[element]
    h = element.__module__ + "@" + element.__qualname__
    for e in sorted (dir (element)):
        entry = getattr (element, e, None)
        if callable (entry):
            h = h + "," + str (_hashFunction (entry))
        else:
            h = h + "/" + str (_hashArgument (entry)) 
    value = _hashString (h)
    _hashClassCache[element] = value
    return value




_hashFunctionCache = {}
def _hashFunction (element):
    """
    Generate hash for a function, returns an unsigned 64 bit number.
    Hashes of classes will track also code changes.
    """
    if element in _hashFunctionCache:
        return _hashFunctionCache[element]
    h = "method@" + element.__qualname__
    code = getattr (element, '__code__', None)
    if code != None:
        h = h + "," + str (_hashString (str (code.co_code))) + "->" + str (_hashArgument (code.co_consts))
    value = _hashString (h)
    _hashFunctionCache[element] = value
    return value




def _hashArgument(element):
    """
    Compute hash code for a single argument, returns an unsigned 64 bit number.
    """
    if element == None:
        return 8726341235
    elif inspect.isclass (element):
        return _hashClass (element)
    elif callable (element):
        # TODO: need refinements for lamda function?
        return _hashFunction (element)
    elif isinstance (element, enum.Enum):
        return _hashString ("e@" + element.__class__.__qualname__ + "@" + element.name + "=" + str (element.value))
    elif isinstance (element, int):
        return _hashString ("i@" + str (element))
    elif isinstance (element, bool):
        return 298374234 if element else -34283749872
    elif isinstance (element, float):
        return _hashString ("f@" + str (element))
    elif isinstance (element, str):
        return _hashString (element)
    elif isinstance (element, (tuple, list)):
        return _hashList (element)
    elif isinstance (element, dict):
        return _hashDict (element)
    elif isinstance (element, object):
        return _hashObject (element)
    assert False, "No rule to hash element."



import math
from . import tree
from . import boolean
from . import transform
from . import solid
from . import shape
from . import attributeTypes
from .util import cache
from .util import vmath
from .util import hash
from .util import differentialdrawing




def pushTransformsFurtherInside (item):
    """
    Push :ref:`Transforms<Transform>` into :ref:`Extrude<Extrude>` and :ref:`Boolean<Boolean>` tree nodes.
    """
    return _pushTransformsFurtherInside (item, [])




def pushExtrusionsFurtherInside (item):
    """
    Push :ref:`Extrudes<Extrude>` into and :ref:`Boolean<Boolean>` tree nodes.
    """
    return _pushExtrusionsFurtherInside (item, [])




def groupsToUnions (item):
    """
    Transform groups into unions.
    """
    return GroupsToUnions.execute (item, tree.Attributes ())




def removeEmpty (item):
    """
    Remove empty items from tree.
    """
    return RemoveEmptyNodes.execute (item, tree.Attributes ())




def defaultColorGenerator (count):
    """
    Generate color for item.
    """
    rsx = 1.0 / 3.0
    if count < 2:
        # TODO: return default color
        return ((1, 0, 0),)
    colors = []
    for index in range (0, count):
        rsp = index / count
        if rsp <= rsx:
            # in red to green range
            ix = (rsp) * 3
            c1 = vmath.vsMulScalar ((1, 0, 0), 1 - ix)
            c2 = vmath.vsMulScalar ((0, 1, 0), ix)
            c = vmath.vsAdd (c1, c2)
        elif rsp <= (2 * rsx):
            # in green to blue range
            ix = (rsp - rsx) * 3
            c1 = vmath.vsMulScalar ((0, 1, 0), 1 - ix)
            c2 = vmath.vsMulScalar ((0, 0, 1), ix)
            c = vmath.vsAdd (c1, c2)
        else:
            # in blue to red range
            ix = (rsp - (2 * rsx)) * 3
            c1 = vmath.vsMulScalar ((0, 0, 1), 1 - ix)
            c2 = vmath.vsMulScalar ((1, 0, 0), ix)
            c = vmath.vsAdd (c1, c2)
        colors.append (c)
    return colors




def distributeColors (item, override: bool = False, colorGenerator = defaultColorGenerator):
    """
    Distribute colors of renderable items.
    """
    colorCount = _DistributeColorsSingleton.countColors (item, override)
    colorMap = colorGenerator (colorCount)
    return _DistributeColorsSingleton.execute (item, override, colorMap)


# === implementation ===


class _DistributeColors (tree.Visitor):
    """
    Distribute colors to renderable items
    """
    def __init__ (self):
        super ().__init__ ()
        self.rule (shape.Shape, leave = self.leaveRenderable)
        self.rule (solid.Solid, leave = self.leaveRenderable)


    class Context (tree.Visitor.Context):
        """
        Visitor for assigning rendering colors.
        """
        def __init__ (self, attributes):
            super ().__init__ (attributes)
            self.contextCollectStage = True
            self.colorMap = None
            self.contextColorId = 0


    def countColors (self, item, override):
        """
        Count colores required for painting items.
        """
        context = _DistributeColors.Context (None)
        self.visit (item, context)
        return context.contextColorId


    def execute (self, item, override, colorMap):
        """
        Run replace node on a tree.
        """
        context = _DistributeColors.Context (None)
        context.colorMap = colorMap
        context.contextCollectStage = False
        return self.visit (item, context)


    def leaveRenderable (self, item, context):
        if context.contextCollectStage:
            if item.attributes.get ('material') == None:
                context.contextColorId += 1
        else:
            if item.attributes.get ('material') == None:
                cid = context.contextColorId
                cid = cid % len (context.colorMap)
                newColor = context.colorMap[cid]
                context.contextColorId += 1
                newMaterialColor = attributeTypes.Material (newColor)
                nAttributes = item.attributes.override ({'material': newMaterialColor})
                return item.copy (attributes = nAttributes)
            return item
        return item

"""
Single instance of distribute colors arlgorithm.
"""
_DistributeColorsSingleton = _DistributeColors ()




def _pushTransformsFurtherInside (item, transformStack):
    """
    Move transforms further inside the tree.
    """
    if item == None:
        return None
    if isinstance (item, tree.Empty):
        return item
    if isinstance (item, transform.Transform):
        transformStack.append (item)
        assert len (item.children) == 1
        result = _pushTransformsFurtherInside (item.children[0], transformStack)
        transformStack.pop ()
        return result
    elif isinstance (item, (boolean.Boolean, solid.LinearExtrude)):
        if len (transformStack) == 0:
            return item
        newChildren = []
        for childId in range (0, len (item.children)):
            newChildren.append (_pushTransformsFurtherInside (item.children[childId], transformStack))
        return item.copy (children = newChildren)
    else:
        if len (transformStack) == 0:
            return item
        else:
            transformed = pushTransformsFurtherInside (item)
            for transformId in range (len (transformStack) - 1, -1, -1):
                transformed = transformStack[transformId].copy (children = (transformed,))
            return transformed




def _pushExtrusionsFurtherInside (item, extrudeStack):
    """
    Move extrusions further inside the tree.
    """
    if item == None:
        return None
    if isinstance (item, tree.Empty):
        return item
    if isinstance (item, (solid.LinearExtrude, solid.RotateExtrude)):
        extrudeStack.append (item)
        assert len (item.children) == 1
        result = _pushExtrusionsFurtherInside (item.children[0], extrudeStack)
        extrudeStack.pop ()
        return result
    elif isinstance (item, boolean.Boolean):
        if len (extrudeStack) == 0:
            return item
        newChildren = []
        for childId in range (0, len (item.children)):
            newChildren.append (_pushExtrusionsFurtherInside (item.children[childId], extrudeStack))
        return item.copy (children = newChildren)
    else:
        if len (extrudeStack) == 0:
            return item
        else:
            extruded = pushExtrusionsFurtherInside (item)
            for extrudedId in range (len (extrudeStack) - 1, -1, -1):
                extruded = extrudeStack[extrudedId].copy (children = (extruded,))
            return extruded




class RemoveNodes (tree.Visitor):
    """
    Remove nodes from csg tree by custom rule.
    """
    def __init__ (self):
        super ().__init__ ()


    def execute (self, item, attributes):
        """
        Run replace node on a tree.
        """
        context = tree.Visitor.Context (attributes)
        result = self.visit (item, context)
        if isinstance (result, tree.Empty):
            return None
        return result


    def enterFilter (self, item, context):
        """
        Enter node filtered out.
        """
        return False


    def leaveFilter (self, item, context):
        """
        Enter node filtered out.
        """
        return tree.Empty ()




class _GroupsToUnions (tree.Visitor):
    """
    Replace group by union nodes.
    """
    def __init__ (self):
        super ().__init__ ()
        self.rule (tree.Group, leave = self.leaveGroup)


    def execute (self, item, attributes):
        """
        Return tree with unions instead of groups.
        """
        context = tree.Visitor.Context (attributes)
        result = self.visit (item, context)
        return result


    def leaveGroup (self, item, context):
        """
        Process group.
        """
        return boolean.Union (children = item.children, name = item.name, attributes = item.attributes)

"""
Single instance of group to union filter.
"""
GroupsToUnions = _GroupsToUnions ()




class _CollectParts (tree.Visitor):
    """
    Collect all parts of a scene.
    """
    def __init__ (self):
        super ().__init__ ()
        self.rule (tree.Part, enter = self.enterPart)


    def execute (self, item, attributes):
        """
        Collect parts from a csg tree. Returns a list of all parts found inside a design.
        """
        context = tree.Visitor.Context (attributes)
        context.__setattr__ ('parts', [])
        self.visit (item, context)
        return context.__getattribute__ ('parts')


    def enterPart (self, item, context):
        """
        Enter part node.
        """
        context.parts.append (item)

        # Do not recurse on part.
        return False

"""
Single instance of part collector.
"""
CollectParts = _CollectParts ()




class _RemoveEmptyNodes (tree.Visitor):
    """
    Remove emtpy nodes from csg tree.
    """
    def __init__ (self):
        super ().__init__ ()

        # deep inspection rules
        self.rule (solid.LinearExtrude, leave = self._leaveSingleChild)
        self.rule (solid.RotateExtrude, leave = self._leaveSingleChild)
        self.rule (shape.Projection, leave = self._leaveSingleChild)
        self.rule (transform.Translate, leave = self._leaveSingleChild)
        self.rule (transform.Scale, leave = self._leaveSingleChild)
        self.rule (transform.Rotate, leave = self._leaveSingleChild)
        self.rule (boolean.Union, leave = self._leaveUnion)
        self.rule (boolean.Difference, leave = self._leaveDifference)
        self.rule (boolean.Intersection, leave = self._leaveIntersection)
        self.rule (tree.Group, leave = self._leaveGroup)


    def execute (self, item, attributes):
        """
        Run replace node on a tree.
        """
        context = tree.Visitor.Context (attributes)
        return self.visit (item, context)


    def _leaveSingleChild (self, item, context):
        """
        Process a tree item with a single child.
        """
        if (item.children[0] == None) or (isinstance (item.children[0], tree.Empty)):
            return item.Empty
        return item


    def _leaveUnion (self, item, context):
        """
        Remove empty items from a union.
        """
        childCount = 0
        children = []
        for child in item.children:
            if child != None:
                if not isinstance (child, tree.Empty):
                    childCount += 1
                    children.append (child)
        if childCount == 0:
            return tree.Empty (name = item.name, attributes = item.attributes)
        if childCount != len (item.children):
            if childCount == 1:
                return children[0]
            return boolean.Union (children = children, name = item.name, attributes = item.attributes)
        else:
            return item


    def _leaveDifference (self, item, context):
        """
        Remove empty items from a difference.
        """
        c1 = item.children[0] if len (item.children) > 0 else None
        c2 = item.children[1] if len (item.children) > 1 else None
        if isinstance (c1, tree.Empty):
            c1 = None
        if isinstance (c2, tree.Empty):
            c2 = None
        if c1 != None:
            if c2 == None:
                return c1
            else:
                return item
        else:
            return tree.Empty (name = item.name, attributes = item.attributes)
        

    def _leaveIntersection (self, item, context):
        """
        Remove empty items from an intersection.
        """
        c1 = item.children[0] if len (item.children) > 0 else None
        c2 = item.children[1] if len (item.children) > 1 else None
        if isinstance (c1, tree.Empty):
            c1 = None
        if isinstance (c2, tree.Empty):
            c2 = None
        if c1 != None:
            if c2 == None:
                return c1
            else:
                return item
        else:
            if c2 == None:
                return tree.Empty (name = item.name, attributes = item.attributes)
            else:
                return c2
        

    def _leaveGroup (self, item, context):
        """
        Remove empty items from a group.
        """
        childCount = 0
        children = []
        for child in item.children:
            if child != None:
                if not isinstance (child, tree.Empty):
                    childCount += 1
                    children.append (child)
        if childCount == 0:
            return tree.Empty (name = item.name, attributes = item.attributes)
        if childCount != len (item.children):
            if childCount == 1:
                return children[0]
            return tree.Group (children = children, name = item.name, attributes = item.attributes)
        else:
            return item
        
"""
Single instance of empty nodes removal filter.
"""
RemoveEmptyNodes = _RemoveEmptyNodes ()




# import open scad tools
from .util.external import openscad




class BezierDDAElement (differentialdrawing.Element):
    """
    Differential drawing element for bezier curve path.
    """
    def __init__ (self, pathPoints):
        super ().__init__ ()
        self.pathPoints = pathPoints

    def pointAt (self, t):
        """
        Calculate a point on the bezier curve.
        """
        return vmath.bezier (self.pathPoints, t)




def _bezierShapeToPolygon (item, rasterizingAttributes):
    """
    Convert a bezier shape to a Polygon.
    """

    # get rasterizing settings
    minAngle, minSize, fixedCount, minSlices = rasterizingAttributes.rasterizingAttributes ()

    # rasterize bezier shape to polygon
    pathes = []
    for curve in item.curves:
        pathes.append (BezierDDAElement (curve))
    dda = differentialdrawing.Path (pathes)
    return dda.toPolygon (initialSegmentCount = 2, maxStepSize = minSize, name = item.name, attributes = item.attributes)




def __reassemblePath (edgesLeft, normalizedEdges, currentPath):
    """
    Reassembles a path from elements.
    """
    # start on an arbitrary path
    pivot = edgesLeft.pop ()
    alreadyProcessed = set ()
    alreadyProcessed.add (pivot)
    currentPath.append (pivot[0])
    currentPath.append (pivot[1])

    # link paths until no path is left.
    elementFound = True
    while elementFound:
        # find next candidate
        lastPoint = currentPath[-1]
        if lastPoint in normalizedEdges:
            nextPoints = normalizedEdges[lastPoint]
            if len (nextPoints) == 1:
                # simple case, got only one candidate
                nextCandidate = (lastPoint, nextPoints[0])
                if lastPoint < nextPoints[0]:
                    nextCandidateKey = (lastPoint, nextPoints[0])
                else:
                    nextCandidateKey = (nextPoints[0], lastPoint)
                if nextCandidateKey in alreadyProcessed:
                    nextCandidate = None
            else:
                nextCandidate = None
                nextCandidateKey = None
                for nextPoint in nextPoints:
                    nextCandidateProto = (lastPoint, nextPoint)
                    if lastPoint < nextPoint:
                        nextCandidateProtoKey = (lastPoint, nextPoint)
                    else:
                        nextCandidateProtoKey = (nextPoint, lastPoint)
                    if not nextCandidateProtoKey in alreadyProcessed:
                        nextCandidate = nextCandidateProto
                        nextCandidateKey = nextCandidateProtoKey
                        break
        else:
            assert False, "invalid point index found"
            nextCandidate = None
            nextCandidateKey = None

        elementFound = nextCandidate != None

        # process candidate
        if nextCandidate != None:
            # TODO: remove from edges left
            if nextCandidateKey in edgesLeft:
                edgesLeft.remove (nextCandidateKey)
            alreadyProcessed.add (nextCandidateKey)
            currentPath.append (nextCandidate[1])

    # Assert path is closed, remove closing point
    assert len (currentPath) > 0, "expected path to be not empty."
    assert currentPath[0] == currentPath[-1]
    currentPath.pop ()
    return




def _projectedMeshToPolygon (mesh):
    """
    Transform a projected mesh into a Polygon.
    """
    # foreach face in "bottom layer" (i.e. all z-coordinates < 0):
    # build a map [points of edge -> (faceIndex, pointIndex1, pointIndex2)]
    edgeMap = {}
    for faceId in range (0, len (mesh.faces)):
        face = mesh.faces[faceId]
        faceAccepted = True
        for edgeId in range (0, 3):
            if mesh.points[face[edgeId]][2] >= 0:
                faceAccepted = False
        if faceAccepted:
            for edgeId in range (0, 3):
                pid1 = face[edgeId % 3]
                pid2 = face[(edgeId + 1) % 3]
                pid = str (pid1) + "," + str (pid2) if pid1 < pid2 else str (pid2) + "," + str (pid1)
                if pid in edgeMap:
                    edgeMap[pid].append ((pid1, pid2))
                else:
                    edgeMap[pid] = [(pid1, pid2)]

    # build edge links
    normalizedEdges = {}
    edgesLeft = set ()
    for edge in edgeMap:
        value = edgeMap[edge]
        if len (value) == 1:
            p0 = value[0][0]
            p1 = value[0][1]
            if p1 < p0:
                t = p0
                p0 = p1
                p1 = t
            edgesLeft.add ((p0, p1))
            if p0 in normalizedEdges:
                normalizedEdges[p0].append (p1)
            else:
                normalizedEdges[p0] = [p1]
            if p1 in normalizedEdges:
                normalizedEdges[p1].append (p0)
            else:
                normalizedEdges[p1] = [p0]

    # rebuild pathes from edge links
    pathes = []
    while len (edgesLeft) > 0:
        currentPath = []
        __reassemblePath (edgesLeft, normalizedEdges, currentPath)
        pathes.append (currentPath)

    # separated pathes, recreate points and indices.
    remappedPointIndices = {}
    newPoints = []
    newPathes = []
    for path in pathes:
        newPath = []
        for point in path:
            if point in remappedPointIndices:
                rpid =  remappedPointIndices[point]
            else:
                rpid = len (newPoints)
                newPoints.append ((mesh.points[point][0], mesh.points[point][1]))
                remappedPointIndices[point] = rpid
            newPath.append (rpid)
        newPathes.append (newPath)

    # select outer path
    outerPathIndex = None
    testEdgeIndex = 0
    testsDone = True
    while testsDone:
        testsDone = False
        for outerFaceTestIndex in range (0, len (newPathes)):
            for innerFaceTestIndex in range (0, len (newPathes)):
                if innerFaceTestIndex != outerFaceTestIndex:
                    if testEdgeIndex < len (newPathes[innerFaceTestIndex]):
                        innerPoint = newPoints[newPathes[innerFaceTestIndex][testEdgeIndex]]
                        if _checkIfPointIsInsidePath (newPoints, newPathes[outerFaceTestIndex], innerPoint):
                            outerPathIndex = outerFaceTestIndex
                            break
                        testsDone = True
            if outerPathIndex != None:
                break
        if testsDone == False:
            break
        testEdgeIndex += 1
    if outerPathIndex == None:
        outerPathIndex = 0

    # build cutouts
    cutouts = []
    for pathIndex in range (0, len (newPathes)):
        if pathIndex != outerPathIndex:
            cutouts.append (newPathes [pathIndex])

    # build polygon from pathes
    return shape.Polygon (newPoints, newPathes[outerPathIndex], cutouts, name = mesh.name, attributes = mesh.attributes)




def _checkIfTestpointIntersectsPath (line1, line2, testPoint):
    # sort that l2rel has maximum x
    l1rel = vmath.vsSub (line1, testPoint)
    l2rel = vmath.vsSub (line2, testPoint)
    if l2rel[0] < l1rel[0]:
        t = l1rel
        l1rel = l2rel
        l2rel = t

    # elements wont match?
    if l2rel[0] < 0:
        return False
    if l1rel[0] > 0:
        return False

    # check if is intersecting
    xDist = l2rel[0] - l1rel[0]
    xRelOff = -l1rel[0] / xDist
    yDist = l2rel[1] - l1rel[1]
    yAtX0 = yDist * xRelOff + l1rel[1]
    return yAtX0 > 0




def _checkIfPointIsInsidePath (points, edges, testPoint):
    """
    Check if a point is inside a path
    """
    intersectionCount = 0
    for edgeId1 in range (0, len (edges)):
        edgeId2 = (edgeId1 + 1) % len (edges)
        if _checkIfTestpointIntersectsPath (points[edgeId1], points[edgeId2], testPoint):
            intersectionCount += 1
    return (intersectionCount % 2) == 1




def toPolygons (item: tree.Item, rasterizingAttributes: tree.Attributes):
    """
    Converts a Shape into a set of polygons.
    """
    # complex shape?
    if isinstance (item, shape.Complex):
        return item.rasterize (rasterizingAttributes)

    # convert bezier shape?
    if isinstance (item, shape.Bezier):
        return _bezierShapeToPolygon (item, rasterizingAttributes)

    # convert bezier shape?
    if isinstance (item, shape.Polar):
        return _polarShapeToPolygon (item, rasterizingAttributes)

    # create shapes from projected solid
    itmesh = solid.Polyhedron.fromSolid (solid.LinearExtrude (children = (item,)), rasterizingAttributes)
    polygons = []
    if isinstance (itmesh, tree.Group):
        for child in itmesh.children:
            if not isinstance (child, tree.Empty):
                polygons.append (_projectedMeshToPolygon (child))
    else:
        if not isinstance (itmesh, tree.Empty):
            polygons.append (_projectedMeshToPolygon (itmesh))

    # create return value
    if len (polygons) == 0:
        return tree.Empty (item.name, item.attributes)
    elif len (polygons) == 1:
        return polygons[0]
    else:
        return tree.Group (children = polygons, name = item.name, attributes = item.attributes)




def toPolyhedrons (item: tree.Item, rasterizingAttributes: tree.Attributes):
    """
    Converts a solid into a set of polyhedrons.
    """
    renderedFile = openscad.getMaterialized ('off', item, rasterizingAttributes)
    return solid.Polyhedron.fromFileOFF (renderedFile)



import copy
from . import immutable




class Attributes (immutable.DataObject):
    """
    Attributes are immutable dictionaries describing rendering attributes of a csg :ref:`Item<Item>`.
    """
    def __init__ (self, attributes = None):
        # check type of parameters
        if attributes == None:
            entries = {}
        else:
            if isinstance (attributes, dict):
                entries = copy.copy (attributes)
            elif isinstance (attributes, Attributes):
                entries = attributes.entries
            else:
                assert False, "Parameter attributes need to be an attributs, dictorionary or None instance."

        super ().__init__ ()

        # Dictionary containing all attribute assignments.
        self.attributes = entries


    @staticmethod
    def defaults ():
        """
        Returns a default attribute set.
        """
        return Attributes ({
            # camera view vector:
            #   either 
            #       - a tuple (eyeX, eyeY, eyeZ, centerX, centerY, centerZ)
            #       - a tuple (translateX, translateY, translateZ, rotateX, rotateY, rotateZ, distance)
            'camera.view':                      None,

            # camera projection mode
            #   either
            #       - "perspective"
            #       - "orthogonal"
            #       - None: default == perspective
            'camera.projection':                None,

            # render attributes
            'render.width':                     None,       # int > 0, default: 800
            'render.height':                    None,       # int > 0, default: 600
            'render.quality':                   None,       # int range 1 .. 12, 12 is best quality, default: 1
            'render.antialias':                 None,       # int range 1 .. 12, 12 is best quality, default: 1
            'render.view.axis':                 None,       # boolean, default: False
            'render.view.scales':               None,       # boolean, default: False
            'render.view.crosshairs':           None,       # boolean, default: False
            'render.colorScheme':               None,       # string, default: "PcsgTheme"

            # default attributes for rasterizing:
            'rasterize.minAngle':               1,
            'rasterize.minSize':                0.01,
            'rasterize.fixedCount':             None,
            'rasterize.fixedSlices':            None,

            # default attributes for rendering:
            'material':                         None
        })


    def merge (self, childAttributes):
        """
        Merges this attribute set with the child attributes and retruns a new instance.
        """
        if childAttributes == None:
            return self
        newAttrs = copy.copy (self.attributes)
        for key in childAttributes.attributes:
            newAttrs[key] = childAttributes.attributes[key]
        return Attributes (newAttrs)


    def override (self, assignments: dict):
        """
        Creates a copy of this attributes object with altered values.
        """
        newAttrs = copy.copy (self.attributes)
        for key in assignments:
            newAttrs[key] = assignments[key]
        return Attributes (newAttrs)


    def get (self, key):
        """
        Get attribute by key. Returns None if the attribute is not set.
        """
        if not key in self.attributes:
            return None
        return self.attributes[key]


    def getAttribute (self, key, defaultValue = None, typeAssert = None):
        """
        Reads a value from the attributes map, may returns a default value.
        """
        if key in self.attributes:
            value = self.attributes[key]
            if value == None:
                return defaultValue
            if typeAssert != None:
                assert isinstance (value, typeAssert)
                return value
        else:
            return defaultValue


    def cameraAttributes (self):
        """
        Get camera settings from Attributes object.
        See :ref:`camera attributes<CameraAttributes>` for more details.
        """
        view = self.getAttribute ('camera.view', (0, 0, 100, 0, 0, 0), (tuple, list))
        assert len (view) in (6, 7)
        projection = self.getAttribute ('camera.projection', 'perspective', str)
        assert projection in ('orthogonal', 'perspective')
        return {
            'view':             view,
            'projection':       projection
        }


    def renderAttributes (self):
        """
        Returns the resolved render attributes from an Attributes object.
        See :ref:`render attributes<RenderAttributes>` for more details.
        """
        return {
            'width':            self.getAttribute ('render.width', 800, (int, float)),
            'height':           self.getAttribute ('render.height', 600, (int, float)),
            'quality':          self.getAttribute ('render.quality', 1, int),
            'antialias':        self.getAttribute ('render.antialias', 1, int),
            'view.axis':        self.getAttribute ('render.view.axis', False, bool),
            'view.scales':      self.getAttribute ('render.view.scales', False, bool),
            'view.crosshairs':  self.getAttribute ('render.view.crosshairs', False, bool),
            'colorScheme':      self.getAttribute ('render.colorScheme', "PcsgTheme", str)
        }

    
    def rasterizingAttributes (self):
        """
        Returns the resolved rasterizing attributes as tuple: (minAngle, minSize, fixedCount, minSlices).
        Note: not all values may be returned, None is also acceptable for undefined values.
        See :ref:`rasterizing attributes<RasterizingAttributes>` for more details.
        """
        minAngle = self.attributes['rasterize.minAngle'] if 'rasterize.minAngle' in self.attributes else None
        minSize = self.attributes['rasterize.minSize'] if 'rasterize.minSize' in self.attributes else None
        fixedCount = self.attributes['rasterize.fixedCount'] if 'rasterize.fixedCount' in self.attributes else None
        minSlices = self.attributes['rasterize.fixedSlices'] if 'rasterize.fixedSlices' in self.attributes else None
        if minAngle == None:
            minAngle = 10
        if minSize == None:
            minSize = 0.01
        return (minAngle, minSize, fixedCount, minSlices)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            'attributes',
        )



from . import immutable
from .attributes import Attributes




class Item (immutable.DataObject):
   
    """
    Base class for a csg tree items without children.
    """
    def __init__ (self, 
                    name: str = None,
                    attributes: Attributes = None
        ):

        super ().__init__ ()

        # check parameter types
        assert (name == None) or isinstance (name, str), "Expected parameter name to be a string."

        #: String containing the name of the tree element. Note: the name will not be included in the objects hash and will be ignored on comparision.
        self.name = name

        #: Attributes of the tree element. When a dictionary is supplied, 
        #: the attributes will converted to an :ref:`Attribute<Attributes>` object.
        self.attributes = attributes if isinstance (attributes, Attributes) else Attributes(attributes)

        #: Number of dimensions as integer: 2 for :ref:`Shapes<Shape>`, 3 for :ref:`Solids<Solid>`, None for non renderable items.
        self.dimensions = None


    #: List of attributes provided to copy constructor.
    #: Each inherited class shall define a list of member names which are passed to the constructor on copy construction.
    #: For inheritance support, also base classes are accepted inside the list.
    @staticmethod
    def _copyConstructorAttributes ():
        """
        Returns a list of attributes provided to the copy constructor.
        Each inherited class shall define a list of member names which are passed to the constructor on copy construction.
        #: For inheritance support, also base classes are accepted inside the list.
        """
        return (
            'name',
            'attributes',
        )


    #: List of attributes for hash and comparision functions, when  not defined _copyConstructorAttributes will be used.
    @staticmethod
    def _compareAttributes ():
        return (
            'attributes',
        )




class Empty (Item):
    """
    Empty csg node. Used as placeholder when modifying the tree.
    """
    def __init__ (self, 
                    name: str = None,
                    attributes: Attributes = None
        ):
        super ().__init__ (name = name, attributes = attributes)

        #: Dimensions of emtpy csg :ref:`Item<Item>` is always None.
        self.dimensions = None

    @staticmethod
    def _compareAttributes ():
        return ()




class Node (Item):
    """
    Base class for csg tree elements with children.
    """
    def __init__ (self, 
                    children: list = None,
                    name: str = None, 
                    attributes: Attributes = None
        ):

        super ().__init__ (
            name = name,
            attributes = attributes
        )

        # check parameter types
        assert isinstance (children, (list, tuple)), "Expected parameter children to be a list or tuple"
        for child in children:
            assert isinstance (child, Item), "Expected child element to be a tree node"

        #: Tuple of child :ref:`Items<Item>` attached to this tree element.
        #: When a list is is supplied, it will be converted to a tuple.
        self.children = tuple (children)


    #: List of attributes for copy construction
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            Item,
            'children'
        )




class Group (Node):
    """
    Container for csg tree items.
    """
    def __init__ (self,
                    children: list = None,
                    name: str = None, 
                    attributes: Attributes = None
        ):

        super ().__init__ (
            children = children,
            name = name,
            attributes = attributes
        )

        # calculate dimension count
        calculatedDimensions = None
        if children != None:
            if len (children) > 0:
                calculatedDimensions = children[0].dimensions
        for childIndex in range (1, len (children)):
            if calculatedDimensions != children[childIndex].dimensions:
                calculatedDimensions = None
                break

        #: Number of dimensions as integer: automatically derived from children when all children have the same number of dimensions, None otherwise.
        self.dimensions = calculatedDimensions




class Part (Node):
    """
    A manufacturable Part encloses a Solid. This class is used to mark Solids as 3D printable components.
    Parts with the same *partName* must have the *child*, otherwise the export will fail.
    """
    def __init__ (self,
                    children: list = None,
                    partName: str = None,
                    name: str = None,
                    attributes: Attributes = None
        ):

        if isinstance (children, Item):
            children = (children,)

        super ().__init__ (
            children = children,
            name = name if name != None else partName,
            attributes = attributes
        )

        # assert valid child
        assert self.children != None, "Expected a single solid as child"
        assert len (self.children) == 1, "Expected a single solid as child"
        assert self.children[0].dimensions == 3, "Expected a single solid as child"

        #: Name of the part, used to generate manufacturing files.
        self.partName = partName
        assert isinstance (partName, str), "Expected part name to be a string."

        #: Parts always have 3 dimensions.
        self.dimensions = 3


    #: List of attributes for copy construction
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            Node,
            'partName'
        )




class Visitor:
    """
    A visitor is used to analyze or modify a csg tree.
    """
    def __init__ (self):
        self._rules = []
        self._compiledRules = None


    class Context:
        """
        Visitor context. Keeps track of parents and resolved :ref:`Attributes<Attributes>`.
        """
        def __init__ (self, attributes: Attributes):
            self._initialAttributes = attributes
            
            #: Stack of parent items for current item. The stack also contains the current item.
            self.stack = []

            #: List of resolved attributes. The list elements corresponds to the resolved attributes for each stack entry.
            self.resolvedAttributes = []


    def visit (self, item, context):
        """
        Visit a tree item.
        Returns the same item as called on or a modified copy of the tree.
        """
        # check for valid item
        if item == None:
            return None
        assert isinstance (item, Item), "Expected a csg tree item"

        # push parent and resolved attributes on context stack
        context.stack.append (item)
        parentAttributes = context.resolvedAttributes[-1] if len (context.resolvedAttributes) > 0 else context._initialAttributes
        if parentAttributes == None:
            context.resolvedAttributes.append (item.attributes)
        else:
            context.resolvedAttributes.append (parentAttributes.merge (item.attributes))

        # dispatch node entry
        recurseOnChildren = self._dispatchEnter (item, context)

        # recurse on children
        if recurseOnChildren == True:
            children = getattr (item, 'children', None)
            visitedChildren = []
            if children != None:
                assert isinstance (children, (list, tuple)), "Children are expected to be a tuple or list."
                for child in children:
                    childAfterVisit = self.visit (child, context)
                    if childAfterVisit != None:
                        visitedChildren.append (childAfterVisit)

            # compare children before and after visitation, may need to create a new object copy.
            cOldLen = len (children) if children != None else 0
            cNewLen = len (visitedChildren)
            childrenChanged = cOldLen != cNewLen
            if childrenChanged == False:
                if cOldLen > 0:
                    for childId in range (0, len (children)):
                        if children[childId] != visitedChildren[childId]:
                            childrenChanged = True
                            break
            if childrenChanged == True:
                item = item.copy (children = visitedChildren)
        
        # dipatch node leave
        result = self._dispatchLeave (item, context)

        # remove parent and resolved attributes from context stack
        context.stack.pop ()
        context.resolvedAttributes.pop ()
        return result


    def rule (self, nodeType: type, enter = None, leave = None):
        """
        Registers a rule for matching classes.
        Enter and leave are the function called while visiting elements of the nodeType,
        when a function is not defined, the default functions will be invoked.
        """
        self._rules.append ((nodeType, enter, leave))
        self._compiledRules = None


    def defaultEnter (self, item, context):
        """
        Default enter function, called when no rule matches.
        Enter functions will return true if the children of the current item should be visited.
        """
        return True


    def defaultLeave (self, item, context):
        """
        Default leave function, called when no rule matches.
        A visitor could replace the item by returning an other instance. A visitor could also remove the node by returning None.
        The by default keeps the item by returning it.
        """
        return item
    

    def _dispatchEnter (self, item, context):
        """
        Dispatch entry of item via compiled rule set.
        """
        compiled = self._getCompiledRules ()
        for rule in compiled:
            if (rule[0] == None) or self._isInstance (type(item), rule[0]):
                result = True
                if rule[1] != None:
                    return rule[1] (item, context)
        return self.defaultEnter (item, context)
    

    def _dispatchLeave (self, item, context):
        """
        Dispatch leave of item via compiled rule set.
        """
        compiled = self._getCompiledRules ()
        for rule in compiled:
            if (rule[0] == None) or self._isInstance (type(item), rule[0]):
                result = True
                if rule[2] != None:
                    return rule[2] (item, context)
        return self.defaultLeave (item, context)


    def _getCompiledRules (self):        
        """
        Get compiled rules for visitor.
        """
        if (self._compiledRules):
            return self._compiledRules
        flattenedRules = []
        for rule in self._rules:
            if self._isInstance (rule[0], list):
                for t in rule[0]:
                    flattenedRules.append ([t, rule[1], rule[2]])
            else:
                flattenedRules.append (rule)
        self._compiledRules = []
        while len (flattenedRules) > 0:
            self._compiledRules.append (self._compileElement (flattenedRules, 0))
        return self._compiledRules


    def _compileElement (self, rulesLeft, index):
        """
        Pick one rule element in order from rulesLeft list.
        """
        if index >= len (rulesLeft):
            return None
        pivot = rulesLeft[index]
        betterIndex = None
        for i in range (index + 1, len (rulesLeft)):
            if self._compileElementCompare (pivot, rulesLeft[i]):
                betterIndex = i
                break
        if betterIndex == None:
            return rulesLeft.pop (index)
        else:
            return self._compileElement (rulesLeft, betterIndex)


    def _compileElementCompare (self, left, right):
        """
        Check if right element class is derived from left one.
        """
        if left[0] == None:
            return True
        return self._isInstance (right[0], left[0])


    def _isInstance (self, clz, needle):
        """
        Check if a class inherits an other one recursively.
        """
        if clz == needle:
            return True
        for base in clz.__bases__:
            if self._isInstance (base, needle):
                return True
        return False



import math
from . import vmath
from .. import util
from ..attributes import Attributes
from .. import shape




class Segment:
    """
    A differential drawing Segment maps a value **t** in range *0 .. 1* to a two dimensional position vector.
    The segment shall be a continiously differentiable curve.
    """
    def __init (self):
        pass


    def pointAt(self, t):
        """
        Calculates a point for t [0..1].
        """
        assert False, "To be implemented by child"




class Element (Segment):
    """
    A differential drawing element maps a function with a ramge *minT .. maxT* to a two dimensional position vector.
    The segment shall be a continiously differentiable curve.
    """
    def __init__(self, minT = 0, maxT = 1, func = None):
        super ().__init__ ()

        #: Mininal **t** passed to function.
        self.minT = minT

        #: Maximal **t** passed to function.
        self.maxT = maxT

        #: Function to calculate a two dimensional position vector for **t**.
        self.func = func
        pass

    def pointAt(self, t):
        """
        Calculates a point for t [0..1] mapped to range [minT..maxT].
        """
        mt = self.minT + t * (self.maxT - self.minT)
        return self._calculate (mt)

    def _calculate(self, mt):
        """
        Calculates a point for t [minT .. maxT].
        """
        if self.func != None:
            return self.func (mt)
        assert False, "To be implemented by child class, optional use func= parameter"




class Path:
    """
    Path consisting of a list of Segments.
    """
    def __init__(self, segments: list = None):
        if segments == None:
            self.elements = []
        else:
            self.elements = segments


    def appendSegment (self, segment):
        """
        Appends a segment to this Path.
        """
        self.elements.append (segment)


    def pointAt(self, t):
        """
        Calculates a point for t [0..1].
        """
        assert len (self.elements) > 0, "no elements in list"
        ta = t * len (self.elements)
        ti = math.floor (ta)
        if ti >= len (self.elements):
            e = self.elements[-1]
            to = 1
        else:
            e = self.elements[ti]
            to = ta - ti
        return e.pointAt (to)


    @staticmethod
    def _append (points, point):
        """
        Append a point to the list of points, skip if duplicated
        """
        if len (points) > 0:
            if (point[0] == points[-1][0]) or (point[1] == points[-1][1]):
                return
        points.append (point)


    def _rasterizeLinearSegment (self, points, element, minT, maxT, minPos, maxPos, maxError):
        """
        Rasterize a segment of an element
        """
        centerT = (minT + maxT) / 2
        centerPoint = element.pointAt (centerT)
        centerOfLine = ((minPos[0] + maxPos[0]) / 2, (minPos[1] + maxPos[1]) / 2)
        dx = centerPoint[0] - centerOfLine[0]
        dy = centerPoint[1] - centerOfLine[1]
        dist = math.sqrt (dx * dx + dy * dy)
        if dist > (maxError / 2):
            self._rasterizeLinearSegment (points, element, minT, centerT, minPos, centerPoint, maxError)
            self._rasterizeLinearSegment (points, element, centerT, maxT, centerPoint, maxPos, maxError)
        else:
            Path._append (points, centerPoint)


    def _rasterize (self, initialSegmentCount = 10, maxStepSize = 0.1):
        """
        Rasterize elements to a path of point coordinates.
        """
        points = []
        for ei in range (0, len (self.elements)):
            element = self.elements[ei]
            initialStepSize = 1.0 / (initialSegmentCount)
            t = 0
            for step in range (0, initialSegmentCount):
                minPos = element.pointAt (t)
                maxPos = element.pointAt (t + initialStepSize)
                Path._append (points, minPos)
                self._rasterizeLinearSegment (points, element, t, t + initialStepSize, minPos, maxPos, maxStepSize)
                Path._append (points, maxPos)
                t += initialStepSize
        return points


    def toPolygon (self, initialSegmentCount = 10, maxStepSize = 0.1, name: str = None, attributes: Attributes = None):
        """
        Create a Polygon from Path.
        """
        points = self._rasterize (initialSegmentCount = initialSegmentCount, maxStepSize = maxStepSize)
        return shape.Polygon (points = points, name = name, attributes = attributes)



    def toBezier (self, initialSegmentCount = 10, maxStepSize = 0.1, name: str = None, attributes: Attributes = None):
        """
        Create a bezier shape from Path.
        """
        assert False, "TODO:"


    def _rasterizeBezier (self, initialSegmentCount = 10, maxStepSize = 0.1, close = True):
        """
        Rasterize elements to a set of bezier curves.
        """
        points = []
        for ei in range (0, len (self.elements)):
            element = self.elements[ei]
            initialStepSize = 1.0 / (initialSegmentCount)
            t = 0
            # create closing curve between last curve element and current curve element
            if len (points) > 0:
                closingNextCurve = []
                self._rasterizeBezierSegment (closingNextCurve, element, 0, 1, maxStepSize, True)
                closingBezier = Path._createConnectionBezier (points, closingNextCurve)
                if closingBezier != None:
                    points = points + closingBezier
                    pass
            for step in range (0, initialSegmentCount):
                # create bezier segments from path
                self._rasterizeBezierSegment (points, element, t, t + initialStepSize, maxStepSize)
                t += initialStepSize
        if close:
            # close bezier curve if requested
            closingBezier = Path._createConnectionBezier (points, points)
            if closingBezier != None:
                points = points + closingBezier
        return points


    @staticmethod
    def _createConnectionBezier (points1, points2, points1Index = None, points2Index = None):
        """
        Create connection between two bezier curves.
        """
        # calculate indices of points to use
        if points1Index == None:
            p1Index = len (points1) - 4
        else:
            p1Index = points1Index
        if points2Index == None:
            p2Index = 0
        else:
            p2Index = points2Index

        # check if we need a closing curve
        startPoint = points1[p1Index + 3]
        endPoint = points2[p2Index]
        seDist = vmath.vsLength (vmath.vsSub (startPoint, endPoint))
        if seDist == 0:
            return None

        # create smooth closing curve
        lastCp = points1[p1Index + 2]
        nextCp = points2[p2Index + 1]
        cp1Direction = vmath.vsSub (lastCp, startPoint)
        cp2Direction = vmath.vsSub (nextCp, endPoint)
        cpd1 = vmath.vsLength (cp1Direction)
        cpd2 = vmath.vsLength (cp2Direction)
        cpFactor = 0.5
        if cpd1 > 0:
            cp1Direction = vmath.vsMulScalar (cp1Direction, (seDist / cpd1) * cpFactor)
        if cpd2 > 0:
            cp2Direction = vmath.vsMulScalar (cp2Direction, (seDist / cpd2) * cpFactor)
        cp1 = vmath.vsAdd (startPoint, cp1Direction)
        cp2 = vmath.vsAdd (endPoint, cp2Direction)
        return [startPoint, cp1, cp2, endPoint]


    @staticmethod
    def _rasterizeBezierSegmentFindScale (startPoint, control1Direction, control2Direction, endPoint, elementCenter, minScale, maxScale):
        # abort recursion
        centerScale = (minScale + maxScale) / 2
        if (maxScale - minScale) < 0.0000001:
            return centerScale
        leftHalfScale = (minScale + centerScale) / 2
        errLeftHalf = vmath.vsLength (vmath.vsSub (elementCenter, vmath.bezier (
            (
                startPoint, 
                vmath.vsAdd (startPoint, vmath.vsMulScalar (control1Direction, leftHalfScale)), 
                vmath.vsAdd (endPoint, vmath.vsMulScalar (control2Direction, leftHalfScale)), 
                endPoint
            ), 0.5
        )))
        rightHalfScale = (centerScale + maxScale) / 2
        errRightHalf = vmath.vsLength (vmath.vsSub (elementCenter, vmath.bezier (
            (
                startPoint, 
                vmath.vsAdd (startPoint, vmath.vsMulScalar (control1Direction, rightHalfScale)), 
                vmath.vsAdd (endPoint, vmath.vsMulScalar (control2Direction, rightHalfScale)), 
                endPoint
            ), 0.5
        )))
        if errLeftHalf < errRightHalf:
            return Path._rasterizeBezierSegmentFindScale (startPoint, control1Direction, control2Direction, endPoint, elementCenter, minScale, centerScale)
        else:
            return Path._rasterizeBezierSegmentFindScale (startPoint, control1Direction, control2Direction, endPoint, elementCenter, centerScale, maxScale)


    @staticmethod
    def _findCurvePointNearestToBezier(bezierPoint, element, minT, maxT):
        centerT = (minT + maxT) / 2
        if (maxT - minT) < 0.0000001:
            return centerT
        
        leftHalfScale = (minT + centerT) / 2
        errLeftHalf = vmath.vsLength (vmath.vsSub (bezierPoint, element.pointAt (leftHalfScale)))
        rightHalfScale = (centerT + maxT) / 2
        errRightHalf = vmath.vsLength (vmath.vsSub (bezierPoint, element.pointAt (rightHalfScale)))

        if errLeftHalf < errRightHalf:
            return Path._findCurvePointNearestToBezier (bezierPoint, element, minT, centerT)
        else:
            return Path._findCurvePointNearestToBezier (bezierPoint, element, centerT, maxT)


    def _rasterizeBezierSegment (self, points, element, minT, maxT, maxError, noSplit = False):
        minDelta = 0.000000001
        angleDelta = 0.0000000001
        dist = maxT - minT
        centerT = (minT + maxT) / 2
        if dist < minDelta:
            # TODO: do not split segment anymore
            noSplit = True

        # create start and end points
        startPoint = element.pointAt (minT)
        endPoint = element.pointAt (maxT)
        seDist = vmath.vsLength (vmath.vsSub (startPoint, endPoint))
        
        # calculate control point distances to match the angles at the end points and scale
        if minT < angleDelta:
            cp1Direction = vmath.vsSub (element.pointAt (minT + angleDelta), startPoint)
        else:
            cp1Direction = vmath.vsSub (element.pointAt (minT + angleDelta), element.pointAt (minT - angleDelta))
        if maxT + angleDelta > 1.0:
            cp2Direction = vmath.vsSub (element.pointAt (maxT - angleDelta), endPoint)
        else:
            cp2Direction = vmath.vsSub (element.pointAt (maxT - angleDelta), element.pointAt (maxT + angleDelta))

        # scale control point directions
        cpd1 = vmath.vsLength (cp1Direction)
        cpd2 = vmath.vsLength (cp2Direction)
        assert cpd1 > 0
        assert cpd2 > 0
        cp1Direction = vmath.vsMulScalar (cp1Direction, (seDist / cpd1))
        cp2Direction = vmath.vsMulScalar (cp2Direction, (seDist / cpd2))
        scale = Path._rasterizeBezierSegmentFindScale (startPoint, cp1Direction, cp2Direction, endPoint, element.pointAt (centerT), 0, 1)
        cp1Direction = vmath.vsMulScalar (cp1Direction, scale)
        cp2Direction = vmath.vsMulScalar (cp2Direction, scale)
        controlPoint1 = vmath.vsAdd (startPoint, cp1Direction)
        controlPoint2 = vmath.vsAdd (endPoint, cp2Direction)

        # determine distance at 0.5, 0.25 and 0.75 ...
        if not noSplit:
            bezierCurve = (startPoint, controlPoint1, controlPoint2, endPoint)
            noSplit = True
            # check if we could match the center within accuracy limit
            if vmath.vsLength (vmath.vsSub (vmath.bezier (bezierCurve, 0.5), element.pointAt (centerT))) > maxError:
                noSplit = False        
            # check if we could match the left half within accuracy limit
            elif vmath.vsLength (vmath.vsSub (vmath.bezier (bezierCurve, 0.25), element.pointAt (Path._findCurvePointNearestToBezier(vmath.bezier (bezierCurve, 0.25), element, minT, maxT)))) > maxError:
                noSplit = False        
            # check if we could match the right half within accuracy limit
            elif vmath.vsLength (vmath.vsSub (vmath.bezier (bezierCurve, 0.75), element.pointAt (Path._findCurvePointNearestToBezier(vmath.bezier (bezierCurve, 0.75), element, minT, maxT)))) > maxError:
                noSplit = False        

        # need to split curves?
        if not noSplit:
                self._rasterizeBezierSegment (points, element, minT, centerT, maxError)                
                self._rasterizeBezierSegment (points, element, centerT, maxT, maxError)                
                return

        # add curve segment
        points.append (startPoint)
        points.append (controlPoint1)
        points.append (controlPoint2)
        points.append (endPoint)



from functools import cmp_to_key




def _optimizeEqualsVector2 (x, y):
    return (x[0] == y[0]) and (x[1] == y[1])




def _optimizeEqualsVector3 (x, y):
    return (x[0] == y[0]) and (x[1] == y[1]) and (x[2] == y[2])




def _optimizeCompareVector2 (x, y):
    if x[0][0] < y[0][0]:
        return -1
    if x[0][0] > y[0][0]:
        return 1
    if x[0][1] < y[0][1]:
        return -1
    if x[0][1] > y[0][1]:
        return 1
    return 1




def _optimizeCompareVector3 (x, y):
    if x[0][0] < y[0][0]:
        return -1
    if x[0][0] > y[0][0]:
        return 1
    if x[0][1] < y[0][1]:
        return -1
    if x[0][1] > y[0][1]:
        return 1
    if x[0][2] < y[0][2]:
        return -1
    if x[0][2] > y[0][2]:
        return 1
    return 1




def _removeDuplicatesVector2 (elements):
    """
    Optimize polyhedron: remove duplicates vom 3D vector list, return new list and remapping dictionary
    """
    if elements == None:
        return (None, None)
    if len (elements) == 0:
        return ((), {})
    orderedElements = sorted (elements, key = cmp_to_key (_optimizeCompareVector2))
    newElements = [orderedElements[0][0]]
    elementMap = {orderedElements[0][1] : 0}
    lastElement = orderedElements[0][0]
    for elementId in range (1, len (orderedElements)):
        e = orderedElements[elementId][0]
        if not _optimizeEqualsVector2 (lastElement, e):
            newElements.append (e)
            lastElement = e
        elementMap[orderedElements[elementId][1]] = len (newElements) - 1
    return (newElements, elementMap)




def _removeDuplicatesVector3 (elements):
    """
    Optimize polyhedron: remove duplicates vom 3D vector list, return new list and remapping dictionary
    """
    if elements == None:
        return (None, None)
    if len (elements) == 0:
        return ((), {})
    orderedElements = sorted (elements, key = cmp_to_key (_optimizeCompareVector3))
    newElements = [orderedElements[0][0]]
    elementMap = {orderedElements[0][1] : 0}
    lastElement = orderedElements[0][0]
    for elementId in range (1, len (orderedElements)):
        e = orderedElements[elementId][0]
        if not _optimizeEqualsVector3 (lastElement, e):
            newElements.append (e)
            lastElement = e
        elementMap[orderedElements[elementId][1]] = len (newElements) - 1
    return (newElements, elementMap)




def _createIndexedList (elements):
    if elements == None:
        return None
    indexed = []
    for index in range (0, len (elements)):
        indexed.append ((elements[index], index))
    return indexed



import math
import inspect
from enum import Enum
from . import tree
from .attributes import Attributes
from .util import cache
from .util import hash




class Shape:
    """
    Informational base class of :ref:`Shape<Shape>` objects. This empty class is implemented for type matching purposes.
    """
    def __init__ (self):
        #: Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2




class Circle(tree.Item, Shape):
    """
    A Circle is a basic :ref:`Shape<Shape>` class.
    
    The Circle can be definied with the parameter *radius* or *diameter*. Only the radius will be stored.
    The Circles center is at coordinate (0, 0).
    """
    def __init__ (self,
                    radius: float = None,
                    diameter: float = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check for valid arguments
        assert (radius != None) or (diameter != None), "Expected parameter radius or diameter"
        assert (radius == None) or (diameter == None), "Parameters radius and diameter can not be used together"
        if diameter != None:
            radius = diameter / 2

        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2

        #: Radius of the circle
        self.radius = radius

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'radius'
        )




class Square(tree.Item, Shape):
    """
    A Square is a basic :ref:`Shape<Shape>` class.
    
    The Square is defined by a two dimensional size tuple.
    The Squares center is at coordinate (0, 0).
    """
    def __init__ (self,
                    size: tuple = (1, 1),
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check for valid size parameter
        assert size != None, "Expected parameter size to be defined"
        if isinstance (size, (int, float)):
            size = (size, size)
        elif isinstance (size, list):
            size = tuple (size)
        elif not isinstance (size, tuple):
            assert False, "Expected size to be a number or a 2 dimensional vector"
        assert len (size) == 2, "Expected size to be a number or a 2 dimensional vector"
        assert isinstance (size[0], (int, float)), "Expected size to be a tuple containing 2 numbers"
        assert isinstance (size[1], (int, float)), "Expected size to be a tuple containing 2 numbers"

        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2

        #: Size of Square as tuple containing (width: float, height: float).
        self.size = size

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'size'
        )
    



class PolygonInterpolation (Enum):
    """
    .. Interpolation type for :ref:`Polygon<Polygon>` rendering.
    """
    #: The polygon will be redered by connecting all points by straight lines.
    Linear = 1

    #: The polygon will be redered by a quadratic interpolation. For calculating the interpolated curve the following point will be taken into account.
    Quadratic = 2

    #: The polygon will be redered by a cubic interpolation. For calculating the interpolated curve the preceeding and following point will be taken into account.
    Cubic = 3




class Polygon(tree.Item, Shape):
    """
    A Polygon is a basic :ref:`Shape<Shape>` class. 

    The Polygon is defined by a list of point forming a closed shape.
    """
    def __init__ (self,
                    points: object = None,
                    outer: list = None,
                    cutouts: list = None,
                    interpolation: PolygonInterpolation = PolygonInterpolation.Linear,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check for valid points
        assert isinstance (points, (list, tuple)), "Expected parameter points to be a tuple of two dimensional vectors"
        newList = []
        for point in points:
            assert isinstance (point, (list, tuple)), "Expected a point to be a list or tuple-"
            assert len (point) == 2, "Expected a point to have 2 dimensions."
            newList.append (tuple (point))
        points = tuple (newList)

        # check outer shape parameter
        if outer == None:
            assert cutouts == None, "Can not use cutouts when no outer shape is specified."
            outer = tuple (range (0, len (points)))
            cutouts = ()
        else:
            # assert points are valid
            assert isinstance (outer, (list, tuple)), "Outer shape must be a list of point indices."
            for point in outer:
                assert isinstance (point, int), "Expected outer shape to consist of point indices."
                assert (point >= 0) and (point < len (points)), "Invalid point index in outer shape."
            outer = tuple (outer)

            # parse in parameter cutouts
            if cutouts == None:
                cutouts = ()
            else:
                # process cutouts
                assert isinstance (cutouts, (list, tuple)), "Expected cutouts to be a list."
                immutableCutouts = []
                for cutout in cutouts:
                    assert isinstance (cutout, (list, tuple)), "Expected each cutout to be a list of point indices."
                    for point in cutout:
                        assert isinstance (point, int), "Expected each cutout be be a list of point indices."
                        assert (point >= 0) and (point < len (points)), "Invalid point index in cutout shape."
                    immutableCutouts.append (tuple (cutout))
                cutouts = tuple (immutableCutouts)


        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2

        #: Tuple of two dimensional vectors describing all points of the Polygon.
        self.points = points

        #: Point indices used for outer shape.
        self.outer = outer

        #: Point array of tuples containing point indices used for cut out shapes.
        self.cutouts = cutouts

        #: Interpolation type used for rendering.
        self.interpolation = interpolation


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'points',
            'interpolation',
            'cutouts',
            'outer'
        )
    

    @staticmethod
    def fromShape (shape: tree.Item, rasterizingAttributes: Attributes = None, name: str = None, attributes: Attributes = None):
        """
        Create a Polygon by rasterizing a two dimensional csg tree :ref:`Item<Item>`.
        The accuracy for rasterizing is calculated from *rasterizingAttributes*.
        When the shape has disjunct areas, a :ref:`Group<Group>` of Polygons will be returned.
        When the shape has no renderable items, :ref:`Empty<Empty>` will be returned.
        """
        # check if node is already cached
        newHash = hash (shape, "polygon from shape", rasterizingAttributes)
        loaded = cache.load (newHash, "rsh")
        if loaded != None:
            return loaded.copy (name = name, attributes = attributes)
        
        # compute polygon and add to cache
        poly = _toPolygonsWrapper (shape, rasterizingAttributes)
        cache.store (poly, newHash, "rsh")
        return poly.copy (name = name, attributes = attributes)




class Bezier(tree.Item, Shape):
    """
    A Bezier shape is a basic :ref:`Shape<Shape>` class. 

    The Bezier shape is defined by a list of bezier curves forming a closed shape.
    Each bezier curve is a tuple containing the start point, two control points and the end point of the segment.
    """
    def __init__ (self,
                    curves: object = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # check for valid points
        assert isinstance (curves, (list, tuple)), "Expected parameter curves to be a tuple of four 2-dimensional vectors."
        if isinstance (curves, list):
            curves = tuple (curves)

        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2

        #: List containing a tuple of 4 points for each bezier segment. A segment consists of four 2-dimensional points: a start point, two control points and an end point.
        self.curves = curves

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Item,
            'curves'
        )

    @staticmethod
    def _compareEndToStart (p1, p2):
        """
        Compare the end of a path to the start of the next one
        """

    def isClosed (self):
        """
        Returns True when the Bezier shape has a closed outline, False otherwise.
        """
        # already cached?
        if 'isClosed' in self._cache:
            return self._cache['isClosed']

        # check if shape is closed
        closed = True
        if len (self.curves) > 0:
            for cId in range (0, len (self.curves)):
                nId = cId + 1
                if nId == len (self.curves):
                    nId = 0
                if not Bezier._compareEndToStart (self.curves[cId], self.curves[nId]):
                    closed = False
                    break
        else:
            closed = False

        # store to cache and return
        self._cache['isClosed'] = closed
        return closed




class Projection (tree.Node, Shape):
    """
    Create a :ref:`Shape<Shape>` by projection of a :ref:`Solid<Solid>` on the x/y plane.
    """
    def __init__ (self,
                    children: object = None,
                    cut: bool = True,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        # assert child is valid
        if isinstance (children, (list, tuple)):
            assert len (children) == 1, "Expected exactly one child"
            assert isinstance (children[0], tree.Item), "Expected a csg tree item as child"
        else:
            assert isinstance (children, tree.Item), "Expected a csg tree item as child"
            children = (children,)
        assert children[0].dimensions == 3, "Expected a three dimensional child"
    
        super ().__init__ (children = children, name = name, attributes = attributes)
      
        #: Tuple containing a single child. The child must be a three dimensional csg tree item.
        self.children = children

        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2

        #: When cut is True, the intersection of the child object with the x/y plane with z=0 will be calculated.
        #: When cut is False, an orthogonal projection will be calculated.
        self.cut = cut == True

    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            tree.Node,
            'cut'
        )




class Complex (tree.Item, Shape):
    """
    Base class for implementing complex :ref:`Shapes<Shape>`.
    """
    def __init__ (self,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name, attributes)

        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2


    def _toPolygon (self, rasterizingAttributes: tree.Attributes):
        """
        To be implemented by child class: convert a Complex to a Polygon.
        """
        assert False, "To be implemented by child class"


    def _toBezier (self, rasterizingAttributes: tree.Attributes):
        """
        To be implemented by child class optionally: convert a Complex to a Bezier shape.
        This function may return None when a conversion to a Bezier shape is not possible.
        """
        return None


    def rasterize (self, rasterizingAttributes: tree.Attributes, tryAsBezierShape = False):
        """
        When *tryAsBezierShape* is True, the Complex is rendered as Bezier curve if possible, otherwise it will be rendered as a Polygon. 
        """
        # try to load from cache
        if tryAsBezierShape:
            fingerprint = hash ("complex-mayBezier", self)
        else:
            fingerprint = hash ("complex", self)
        cached = cache.load (fingerprint, 'item')
        if cached != None:
            return cached

        # generate rasterized bezier shape
        if tryAsBezierShape:
            generated = self._toBezierShape (rasterizingAttributes)
        else:
            generated = None

        # generate rasterized polygon
        if generated == None:
            generated = self._toPolygon (rasterizingAttributes)

        # store to cache and return
        cache.store (generated, fingerprint, 'item')
        return generated




class Polar (Complex):
    """
    A :ref:`Shape<Shape>` defined by a polar curve function.
    """
    def __init__ (self,
                    func = None,
                    name: str = None,
                    attributes: Attributes = None    
        ):
        super ().__init__ (name = name, attributes = attributes)

        # TODO: check if func parameter is valid
        assert inspect.isfunction (func), "Expected function as source of polar shape."

        # Number of dimensions is always 2 for :ref:`Shapes<Shape>`.
        self.dimensions = 2

        #: Function mapping a parameter phi in range 0 to 2\*pi to a a radius.
        self.func = func


    def _toPolygon (self, rasterizingAttributes: tree.Attributes):
        return _polarToPolygon (self, rasterizingAttributes)


    def _toBezier (self, rasterizingAttributes: tree.Attributes):
        return _polarToBezier (self, rasterizingAttributes)


    #: List of attributes for copy construction.
    @staticmethod
    def _copyConstructorAttributes ():
        return (
            Complex,
            'func'
        )




# import algorithms
from . import algorithms
from .util import differentialdrawing




def _toPolygonsWrapper (item: tree.Item, rasterizingAttributes: tree.Attributes):
    """
    Delegate to algorithms.
    """
    return algorithms.toPolygons (item, rasterizingAttributes)




class PolarDDAElement (differentialdrawing.Element):
    """
    Differential drawing element for polar curves
    """
    def __init__ (self, func):
        super ().__init__ ()
        self.func = func

    def pointAt (self, t):
        """
        Calculate point on polar curve.
        """
        phi = t * 2 * math.pi
        radius = self.func (phi)
        return (
            math.sin (phi) * radius,
            math.cos (phi) * radius
        )




def _polarToPolygon (item: Polar, rasterizingAttributes: tree.Attributes):
    """
    Rasterize Polar as Polygon.
    """
    # get rasterizing settings
    minAngle, minSize, fixedCount, minSlices = rasterizingAttributes.rasterizingAttributes ()

    # rasterize bezier shape to polygon
    dda = differentialdrawing.Path ([PolarDDAElement (item.func)])
    return dda.toPolygon (initialSegmentCount = 2, maxStepSize = minSize, name = item.name, attributes = item.attributes)




def _polarToBezier (item: Polar, rasterizingAttributes: tree.Attributes):
    """
    Rasterize Polar as Bezier curve.
    """
    # get rasterizing settings
    minAngle, minSize, fixedCount, minSlices = rasterizingAttributes.rasterizingAttributes ()

    # rasterize bezier shape to polygon
    dda = differentialdrawing.Path ([PolarDDAElement (item.func)])
    return dda.toBezier (initialSegmentCount = 2, maxStepSize = minSize, name = item.name, attributes = item.attributes)
